# -*- coding:utf-8 -*-

"""Serial agent tests."""

from __future__ import (absolute_import, division, print_function,
                        unicode_literals)
from builtins import *  # pylint:disable=W0401,W0614,W0622


import mock

from iotlabmqtt import serial
from . import TestCaseImproved


class LineHandlerTest(TestCaseImproved):
    """Test LineHandler."""
    def test_line_handler_utf_8(self):
        """Test LineHandler newline in utf-8 sequences."""
        unicode_str = 'abĊde\nfgԊȊjk\nmno'
        lines = unicode_str.splitlines(False)

        self.assertEqual(lines, ['ab\u010Ade', 'fg\u050A\u020Ajk', 'mno'])

        callback = mock.Mock()
        line_handler = serial.LineHandler(callback)

        utf8_str = unicode_str.encode('utf-8')
        # Iter on each bytes
        for single_byte in self._bytes_iter(utf8_str):
            line_handler(single_byte)

        # last line incomplete
        calls = [mock.call(line.encode('utf-8')) for line in lines[:-1]]
        callback.assert_has_calls(calls)
        callback.assert_has_calls([])

    @staticmethod
    def _bytes_iter(bytes_str):
        """Python2 / python3 compatible bytes iterator.

        Python2 iter with bytes, python3 with ints.
        http://stackoverflow.com/a/14267935/395687
        """
        return [bytes_str[i:i + 1] for i in range(len(bytes_str))]



# -*- coding:utf-8 -*-

"""Common functions for iotlabmqtt agents."""


from __future__ import (absolute_import, division, print_function,
                        unicode_literals)
from builtins import *  # pylint:disable=W0401,W0614,W0622


import os
import sys
import hashlib
import signal
import string
import argparse


def topic_lazyformat(topic, **kwargs):
    """Format only replace values in 'kwargs' keeping the other intact.


    >>> topic_lazyformat('a/{b}/c/{d}/e') == 'a/{b}/c/{d}/e'
    True

    >>> topic_lazyformat('a/{b}/c/{d}/e', b='B', any='value') == 'a/B/c/{d}/e'
    True

    >>> topic_lazyformat('a/{b}/c/{d}/e', b='B', d='D') == 'a/B/c/D/e'
    True
    """
    fields = topic_fields(topic)
    # identity formats
    fmt = {f: '{%s}' % f for f in fields}

    # overwrite this ones
    for name, value in kwargs.items():
        fmt[name] = value

    return topic.format(**fmt)


def topic_fields(topic):
    """Extract named fields from topic.

    >>> topic_fields('{archi}/{num}/line/{clientid}/{requestid}')
    ['archi', 'num', 'clientid', 'requestid']

    >>> topic_fields('topic/super/cool')
    []

    # Invalid topics

    >>> topic_fields('a/b/{c}/{c}')
    Traceback (most recent call last):
    ValueError: Named fields should be appear only once

    >>> topic_fields('a/{0}/b')
    Traceback (most recent call last):
    ValueError: Use only simple named fields

    >>> topic_fields('a/{}/b')
    Traceback (most recent call last):
    ValueError: Use only simple named fields

    >>> topic_fields('a/{self.test}/b')
    Traceback (most recent call last):
    ValueError: Use only simple named fields

    >>> topic_fields('a/{first}{second}/b')
    Traceback (most recent call last):
    ValueError: There should be only one named field per level

    >>> topic_fields('{name!r}')
    Traceback (most recent call last):
    ValueError: There should be no format or conversion
    """
    fmt = string.Formatter()

    # Prepend with '/' so all fiels have `text` ending with '/'
    _topic = '/' + topic
    fields = [str(name) for text, name, spec, conv in fmt.parse(_topic)
              if _is_named_field(text, name, spec, conv)]
    if len(fields) != len(set(fields)):
        raise ValueError('Named fields should be appear only once')
    return fields


def _is_named_field(text, name, spec, conversion):
    """Return if current field is a named one."""
    # Not a field, just end of the string
    if name is None:
        return False

    # Invalid formats
    if not name or name.isdigit() or '.' in name:
        raise ValueError('Use only simple named fields')
    if not text.endswith('/'):
        raise ValueError('There should be only one named field per level')
    if spec or conversion:
        raise ValueError('There should be no format or conversion')

    return True


class MQTTAgentArgumentParser(argparse.ArgumentParser):
    """ArgumentParser with common agents arguments."""
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self._add_common_arguments()

    def _add_common_arguments(self):
        """Add common agents arguments to parser."""
        self.add_argument('--prefix', help='Topics prefix', default='')
        self.add_argument('--broker-port', help='Broker port')
        self.add_argument('broker', help='Broker address')

    def add_agenttopic_argument(self):
        """Add common agents arguments to parser."""
        self.add_argument('--agenttopic', help='Agent topic overwrite')


def hostname():
    """Return system 'hostname'.

    Read from environment variable to allow overwriting in tests.
    Read it from os.uname() if not available.
    """
    return os.environ.get('HOSTNAME', os.uname()[1])


def traceback_error():
    """Get error from traceback to return as answer."""
    return '%s' % (sys.exc_info()[1],)


def short_hash(data, algorithm='sha1', length=6):
    """Returns hash of ``length`` char for ``data`` with ``algorithm``.

    >>> short_hash(b'example') == 'c3499c'
    True

    >>> short_hash(b'example', 'sha1', 6) == 'c3499c'
    True
    """
    hash_ = hashlib.new(algorithm)
    hash_.update(data)
    hash_hexstr = hash_.hexdigest()
    return hash_hexstr[:length]


def synchronized(lockname):
    """Method decorator to synchronize method calls.

    Class should define a lock object at attribute named ``SYNCHRONIZEDATTR``

    http://caffeinatedideas.com/2014/12/12/java-synchronized-in-python.html
    """
    import functools

    def _wrapper(method_):
        @functools.wraps(method_)
        def _synchronized_method(self, *args, **kwargs):
            with getattr(self, lockname):
                return method_(self, *args, **kwargs)
        return _synchronized_method
    return _wrapper


def wait_sigint():
    """Pause until Ctrl+C."""
    try:
        signal.pause()
    except KeyboardInterrupt:
        pass



# -*- coding:utf-8 -*-
"""Integration tests for IoT-LAB MQTT radiosniffer client and server."""
from __future__ import (absolute_import, division, print_function,
                        unicode_literals)
from builtins import *  # pylint:disable=W0401,W0614,W0622

import os
import contextlib
try:
    from StringIO import StringIO
except ImportError:
    from io import StringIO

import mock

from iotlabmqtt import node
from iotlabmqtt.clients import node as node_client

from . import IntegrationTestCase
from .. import TestCaseImproved

# pylint:disable=invalid-name


class NodeIntegrationTest(IntegrationTestCase):
    """Test node client and server using a broker."""
    @contextlib.contextmanager
    def start_client_and_server(self, brokerport):
        """Start node client and server context manager.

        Yields client and stdout mock.
        """
        # pylint:disable=attribute-defined-outside-init
        args = ['localhost', '--broker-port', '%s' % brokerport,
                '--prefix', 'node/test/prefix']
        args += ['--experiment-id', '12345',
                 '--iotlab-user', 'us3rn4me',
                 '--iotlab-password', 'p4sswd']
        opts = node.PARSER.parse_args(args)
        with mock.patch('iotlabcli.experiment.get_experiment') as get_exp:
            get_exp.return_value = {'state': 'Running'}
            server = node.MQTTNodeAgent.from_opts_dict(**vars(opts))

        assert server.iotlabapi.expid == 12345
        assert server.iotlabapi.api.auth.username == 'us3rn4me'
        assert server.iotlabapi.api.auth.password == 'p4sswd'
        server.iotlabapi.api = None  # Breaks on calls

        server.start()
        self.server = server

        try:
            args = ['localhost', '--broker-port', '%s' % brokerport,
                    '--prefix', 'node/test/prefix',
                    '--site', server.HOSTNAME]
            opts = node_client.PARSER.parse_args(args)
            client = node_client.NodeShell.from_opts_dict(**vars(opts))
            client.start()
            try:
                with mock.patch('sys.stdout', new_callable=StringIO) as stdout:
                    yield client, stdout
            finally:
                client.stop()
        finally:
            server.stop()

    def test_node_agent_normal_cases(self):
        """Test node agent normal case."""
        with self.start_client_and_server(self.BROKERPORT) as (client, stdout):
            self._test_node_agent_normal_cases(client, stdout)

    @mock.patch('iotlabcli.node.node_command')
    def _test_node_agent_normal_cases(self, client, stdout, node_command=None):
        """Normal cases for node agent."""
        node_host = 'm3-1.%s.iot-lab.info' % self.server.HOSTNAME

        # Reset
        node_command.return_value = {'0': [node_host]}
        client.onecmd('reset m3 1')
        self.assertEqual(stdout.getvalue(), '')

        # Reset with error
        node_command.return_value = {'1': [node_host]}
        client.onecmd('reset m3 1')
        out = 'Execution failed on node\n'
        self.assertEqual(stdout.getvalue(), out)
        stdout.seek(0)
        stdout.truncate(0)

        # Power ON
        node_command.return_value = {'0': [node_host]}
        client.onecmd('poweron m3 1')
        self.assertEqual(stdout.getvalue(), '')

        node_command.return_value = {'1': [node_host]}
        client.onecmd('poweron m3 1')
        out = 'Execution failed on node\n'
        self.assertEqual(stdout.getvalue(), out)
        stdout.seek(0)
        stdout.truncate(0)

        # Power OFF
        node_command.return_value = {'0': [node_host]}
        client.onecmd('poweroff m3 1')
        self.assertEqual(stdout.getvalue(), '')

        node_command.return_value = {'1': [node_host]}
        client.onecmd('poweroff m3 1')
        out = 'Execution failed on node\n'
        self.assertEqual(stdout.getvalue(), out)
        stdout.seek(0)
        stdout.truncate(0)

    def test_node_agent_update(self):
        """Test node agent update with firmware."""
        with self.start_client_and_server(self.BROKERPORT) as (client, stdout):
            self._test_node_agent_update(client, stdout)

    @mock.patch('iotlabcli.node.node_command')
    def _test_node_agent_update(self, client, stdout, node_command=None):
        """Update firmware."""

        node_host = 'm3-1.%s.iot-lab.info' % self.server.HOSTNAME
        firmware_path = self.file_path('m3_firmware.elf')
        firmware_sha1 = '2386ee'

        # pylint:disable=unused-argument
        # pylint:disable=protected-access
        def _node_cmd(api, cmd, exp_id, nodes, fw_path):
            """Got issues when file was not flushed, so verify content."""
            with open(fw_path, mode='rb') as fw_:
                data = fw_.read()
                self.assertEqual(data, b'ELF32')

                sha1 = node.common.short_hash(data)
                self.assertTrue(fw_path.endswith('--sha1:%s' % sha1))
            return mock.DEFAULT
        node_command.side_effect = _node_cmd

        # Update
        node_command.return_value = {'0': [node_host]}
        client.onecmd('update m3 1 %s' % firmware_path)
        self.assertEqual(stdout.getvalue(), '')
        self.assertTrue(node_command.called)

        # Verify call args
        api, cmd, exp_id, nodes, fw_path = node_command.call_args[0]
        self.assertEqual(
            (api, cmd, exp_id, nodes),
            (self.server.iotlabapi.api, 'update', 12345, [node_host]))
        # Verify content with sha1
        self.assertTrue(fw_path.endswith('--sha1:%s' % firmware_sha1))

        # Test error
        node_command.return_value = {'1': [node_host]}
        client.onecmd('update m3 1 %s' % firmware_path)
        out = 'Execution failed on node\n'
        self.assertEqual(stdout.getvalue(), out)
        stdout.seek(0)
        stdout.truncate(0)

    def test_node_agent_update_idle(self):
        """Test node agent update with firmware idle."""
        with self.start_client_and_server(self.BROKERPORT) as (client, stdout):
            self._test_node_agent_update_idle(client, stdout)

    @mock.patch('iotlabcli.node.node_command')
    def _test_node_agent_update_idle(self, client, stdout, node_command=None):
        """Update firmware with idle."""

        node_host = 'm3-1.%s.iot-lab.info' % self.server.HOSTNAME

        # Update
        node_command.return_value = {'0': [node_host]}
        client.onecmd('update m3 1')
        self.assertEqual(stdout.getvalue(), '')
        self.assertTrue(node_command.called)

        # Verify call
        node_command.assert_called_with(
            self.server.iotlabapi.api, 'update-idle', 12345, [node_host], None)

    def test_node_agent_error_cb(self):
        """Test node client error_cb."""
        with self.start_client_and_server(self.BROKERPORT) as (client, stdout):
            self._test_node_agent_error_cb(client, stdout)

    def _test_node_agent_error_cb(self, client, stdout):
        """Call node client error cb."""
        topic = self.server.topics['error'].topic
        topic = topic.replace('/error/', '')
        topic = os.path.join(topic, 'a/topic/on/server')

        error = 'test manual error'.encode('utf-8')
        self.server.topics['error'].publish_error(self.server.client, topic,
                                                  error)

        err = ('NODE ERROR: a/topic/on/server: test manual error\n'
               '(Cmd) ')

        self.assertEqualTimeout(stdout.getvalue, err, 2)
        stdout.seek(0)
        stdout.truncate(0)


@mock.patch('sys.stdout', new_callable=StringIO)
class NodeClientErrorTests(TestCaseImproved):
    """Test Node parsing errors."""
    def setUp(self):
        args = ['localhost', '--broker-port', '%s' % 54321,
                '--prefix', 'node/test/prefix',
                '--site', node.MQTTNodeAgent.HOSTNAME]
        opts = node_client.PARSER.parse_args(args)
        self.client = node_client.NodeShell.from_opts_dict(**vars(opts))
        # Publish should crash
        self.client.publish = None

    def test_reset(self, stdout):
        """Test reset errors."""
        hlp = ('Error: Invalid arguments\n'
               'Usage: reset ARCHI NUM\n'
               '  ARCHI: m3/a8\n'
               '  NUM:   node num\n')

        # Missing port
        self.client.onecmd('reset m3')
        self.assertEqual(stdout.getvalue(), hlp)
        stdout.seek(0)
        stdout.truncate(0)

        # Invalid port
        self.client.onecmd('reset m3 abc')
        self.assertEqual(stdout.getvalue(), hlp)
        stdout.seek(0)
        stdout.truncate(0)

        # Too many arguments
        self.client.onecmd('reset m3 123 anotherone')
        self.assertEqual(stdout.getvalue(), hlp)
        stdout.seek(0)
        stdout.truncate(0)

    def test_poweron(self, stdout):
        """Test poweron errors."""
        hlp = ('Error: Invalid arguments\n'
               'Usage: poweron ARCHI NUM\n'
               '  ARCHI: m3/a8\n'
               '  NUM:   node num\n')

        # Missing port
        self.client.onecmd('poweron m3')
        self.assertEqual(stdout.getvalue(), hlp)
        stdout.seek(0)
        stdout.truncate(0)

        # Invalid port
        self.client.onecmd('poweron m3 abc')
        self.assertEqual(stdout.getvalue(), hlp)
        stdout.seek(0)
        stdout.truncate(0)

        # Too many arguments
        self.client.onecmd('poweron m3 123 anotherone')
        self.assertEqual(stdout.getvalue(), hlp)
        stdout.seek(0)
        stdout.truncate(0)

    def test_poweroff(self, stdout):
        """Test poweroff errors."""
        hlp = ('Error: Invalid arguments\n'
               'Usage: poweroff ARCHI NUM\n'
               '  ARCHI: m3/a8\n'
               '  NUM:   node num\n')

        # Missing port
        self.client.onecmd('poweroff m3')
        self.assertEqual(stdout.getvalue(), hlp)
        stdout.seek(0)
        stdout.truncate(0)

        # Invalid port
        self.client.onecmd('poweroff m3 abc')
        self.assertEqual(stdout.getvalue(), hlp)
        stdout.seek(0)
        stdout.truncate(0)

        # Too many arguments
        self.client.onecmd('poweroff m3 123 anotherone')
        self.assertEqual(stdout.getvalue(), hlp)
        stdout.seek(0)
        stdout.truncate(0)

    def test_update(self, stdout):
        """Test update errors."""
        hlp = ('Error: Invalid arguments\n'
               'Usage: update ARCHI NUM FIRMWAREPATH\n'
               '  ARCHI: m3/a8\n'
               '  NUM:   node num\n'
               '  FIRMWAREPATH: Path to the firmware.\n'
               '                Idle firmware if not provided\n')

        # Missing port
        self.client.onecmd('update m3')
        self.assertEqual(stdout.getvalue(), hlp)
        stdout.seek(0)
        stdout.truncate(0)

        # Invalid port
        self.client.onecmd('update m3 abc')
        self.assertEqual(stdout.getvalue(), hlp)
        stdout.seek(0)
        stdout.truncate(0)

        # Invalid file path
        self.client.onecmd('update m3 1 /non/existing/path/fw.elf')
        err = ("Could not open file: [Errno 2] No such file or directory: "
               "'/non/existing/path/fw.elf'\n")
        self.assertEqual(stdout.getvalue(), err + hlp)
        stdout.seek(0)
        stdout.truncate(0)

        # Too many arguments
        cmd = 'update m3 123 %s anotherone' % os.devnull
        self.client.onecmd(cmd)
        self.assertEqual(stdout.getvalue(), hlp)
        stdout.seek(0)
        stdout.truncate(0)



# -*- coding:utf-8 -*-

"""Common MQTT class for iotlabmqtt agents."""

from __future__ import (absolute_import, division, print_function,
                        unicode_literals)
from builtins import *  # pylint:disable=W0401,W0614,W0622

import re
import os.path
import functools
import contextlib
import threading
import packaging.version

import paho.mqtt
import paho.mqtt.client as mqtt

from . import common

PAHO_VERSION = packaging.version.parse(paho.mqtt.__version__)


class MQTTClient(mqtt.Client):
    """MQTT Agent implementation."""

    SUBSCRIBE_TIMEOUT = 10

    def __init__(self, server, port, topics=None):
        super().__init__()

        self.server = server
        self.port = int(port or 1883)
        self.topics = topics or ()

        self._subscribed = threading.Event()

    def on_connect(self, mqttc, obj, flags, rc):  # pylint:disable=W0221,W0613
        """On connect, subscribe to all topics."""
        self._subscribe_topics()

    def _subscribe_topics(self):
        """Suscribe to topics."""
        subtopics = self._subscribable_topics()
        # No subscribe to do
        if not subtopics:
            self._subscribed.set()
            return

        self._log_sub_topic(subtopics)

        topics = (t.subscribe_topic for t in subtopics)
        topics = (self._paho_topic_python2_3(t) for t in topics)
        topics_list = [(t, 0) for t in topics]

        self._subscribed.clear()
        self.subscribe(topics_list)

    @staticmethod
    def _paho_topic_python2_3(topic):
        """Fix issue with paho 1.2.x and python2.

        It requires a 'str' topic and tries to encode it to 'utf-8' after.
        For python2, force it to be ascii bytes so auto-conversion will work.
        """
        import sys  # pylint:disable=redefined-outer-name
        if sys.version_info[0] >= 3:
            return topic

        # Handle paho-mqtt 1.2.x
        if (PAHO_VERSION >= packaging.version.parse('1.2') or
                PAHO_VERSION < packaging.version.parse('1.3')):
            return topic.encode('ascii')

        # It may change in 1.3, so raise a warning to force to check this
        from warnings import warn
        warn('paho-mqtt version %s has not been validated' % PAHO_VERSION)

        return topic

    def _subscribable_topics(self):
        """Topics that are subscrible."""
        return [t for t in self.topics if t.subscribe_topic is not None]

    @staticmethod
    def _log_sub_topic(topics):
        """Log on which subscribed we are subscribing."""
        for topic in topics:
            print('Subscribing to: %s' % (topic.topic,))

    def on_subscribe(self, mqttc, obj, mid, qos):  # pylint:disable=W0221,W0613
        """Unlock '_subscribed' event."""
        self._subscribed.set()

    # def on_publish(self, mqttc, obj, mid):  #  Ignore

    def on_message(self, mqttc, obj, msg):  # pylint:disable=W0221,W0613
        # This should do an error
        print('on_message(%s): %r' % (msg.topic, msg.payload))

    def on_log(self, mqttc, obj, level, string):  # pylint:disable=W0221,W0613
        if level != 16:
            # This should do an error
            print('on_log(%r): %s' % (level, string))

    def start(self):
        """Start MQTT Agent and subscribe to topics."""
        self._register_topics_callbacks()

        self.connect(self.server, self.port)
        self.loop_start()

        subscribed = self._subscribed.wait(self.SUBSCRIBE_TIMEOUT)
        if not subscribed:
            raise RuntimeError('Topics subscribe timeout')

    def _register_topics_callbacks(self):
        """Register the callbacks for topics."""
        topics = (t for t in self.topics if t.callback is not None)
        for topic in topics:
            self.message_callback_add(topic.subscribe_topic, topic.callback)

    def stop(self):
        """Stop MQTT Agent."""
        self.loop_stop()

    @contextlib.contextmanager
    def message_callback(self, topic, callback):
        """Contextmanager that sets topic callback for current context."""
        self.message_callback_add(topic, callback)
        try:
            yield
        finally:
            self.message_callback_remove(topic)

    def publisher(self, topic):
        """Return a function that publishes on ``topic``."""
        return functools.partial(self.publish, topic)

    def publish(self, topic, payload=None, qos=0, retain=False):
        """Publish but requires strings to be bytes."""
        payload = self._bytes_safe_payload(payload)
        return super().publish(topic, payload=payload, qos=qos, retain=retain)

    @staticmethod
    def _bytes_safe_payload(payload):
        """Convert 'payload' to be a bytearray if type 'bytes'.
        Reject 'str' as it allows not managing encoding.

        paho-mqtt (1.2) does not correctly handles python2 'str'
        (== python3 bytes) and tries to encode them to 'utf-8'.
        Giving a ``bytearray`` circumvents it.
        https://github.com/eclipse/paho.mqtt.python/issues/125
        """
        assert not isinstance(payload, str)
        if isinstance(payload, bytes):
            return bytearray(payload)
        return payload

    @classmethod
    def from_opts_dict(cls, broker, broker_port, **_):
        """Create class from argparse entries."""
        return cls(broker, port=broker_port)


def _fmt_topic(topic, prefix='', static_fmt_dict=None):
    """Prepend prefix to prefix and update values from static_fmt_dict.

    >>> _fmt_topic('a/{b}/c/{d}', 'pfx', {'b': 'B'}) == 'pfx/a/B/c/{d}'
    True
    >>> _fmt_topic('a/{b}/c/{d}') == 'a/{b}/c/{d}'
    True
    """
    static_fmt_dict = static_fmt_dict or {}
    topic = os.path.join(prefix, topic)
    topic = common.topic_lazyformat(topic, **static_fmt_dict)
    return topic


def generate_topics_dict(topics_dict, prefix='', agenttopic='',
                         static_fmt_dict=None):
    """Generate full topics dict and add 'agenttopic'."""

    topics = {k: os.path.join(agenttopic, v) for k, v in topics_dict.items()}
    topics['agenttopic'] = agenttopic

    return format_topics_dict(topics, prefix, static_fmt_dict)


def format_topics_dict(topic_dict, prefix='', static_fmt_dict=None):
    """Return a new dict with formatted topics.

    >>> topics = {'a': 'a/{b}/c/{d}', 'b': '1/2/{three}'}
    >>> ret = {'a': 'pfx/a/B/c/{d}', 'b': 'pfx/1/2/3'}

    >>> format_topics_dict(topics, 'pfx', {'b': 'B', 'three': '3'}) == ret
    True
    """

    formatted_topics = {n: _fmt_topic(t, prefix, static_fmt_dict)
                        for n, t in topic_dict.items()}
    return formatted_topics


class Topic(object):
    """Topic base class."""
    LEVEL = r'(?P<%s>[^/]+)'

    def __init__(self, topic, callback=None):
        self.topic = topic
        self.fields = common.topic_fields(self.topic)
        self.subscribe_topic = self._topic_wildcard(self.topic, *self.fields)
        self.match_re = self._topic_match_re(self.topic, *self.fields)
        self.callback = self.wrap_callback(callback) if callback else None

    def fields_values(self, topic):
        """Extract named fields values from actual topic."""
        match = self.match_re.match(topic)
        fields = {f: match.group(f) for f in self.fields}
        return fields

    def wrap_callback(self, callback):
        """Wrap callback to call it with fields arguments values."""
        @functools.wraps(callback)
        def _wrapper(mqttc, obj, msg):  # pylint:disable=unused-argument
            fields = self.fields_values(msg.topic)
            return callback(msg, **fields)

        return _wrapper

    @staticmethod
    def _topic_wildcard(topic, *fields):
        """Convert `topic` named `fields` to '+' mqtt wildcard."""
        fmts = {f: '+' for f in fields}
        return topic.format(**fmts)

    @classmethod
    def _topic_match_re(cls, topic, *fields):
        """Convert `topic` to a re pattern that extracts fields values."""
        fmts = {f: cls.LEVEL % f for f in fields}
        return re.compile(topic.format(**fmts))


class NullTopic(Topic):
    """NullTopic just to store Topic value."""

    def __init__(self, topic, callback=None):
        assert callback is None
        super().__init__(topic, callback)
        # Disable subscription for NullTopic
        self.subscribe_topic = None


class InputServer(Topic):
    """Topic implementation for an Input Server."""
    pass


class InputClient(NullTopic):
    """Topic implementation for an Input Client."""

    def send(self, client, data, **fmt):
        """Send ``data`` to topic formatted with **fmt."""
        topic = self.topic.format(**fmt)
        return client.publish(topic, data)


class OutputServer(InputClient):
    """Topic implementation for an Output Server."""
    pass


class OutputClient(InputServer):
    """Topic implementation for an Output Client."""
    pass


class LogTopic(Topic):
    """Topic to log all messages."""

    def __init__(self, topic, callback=None):
        logtopic = os.path.join(topic, '#')

        super().__init__(logtopic, callback)


class RequestTopic(object):
    """RequestTopic format."""

    REQUEST_FIELDS = {'clientid', 'requestid'}
    REQUEST_SUFFIX = 'ctl/{command}/request/{clientid}/{requestid}'
    REPLY_SUFFIX = 'ctl/{command}/reply/{clientid}/{requestid}'

    _TOPIC_RE = r'^{}/(request|reply)/{}/{}$'.format(r'(?P<topic>.*)',
                                                     Topic.LEVEL % 'clientid',
                                                     Topic.LEVEL % 'requestid')
    TOPIC_RE = re.compile(_TOPIC_RE)
    TOPIC_REPL_FMT = r'\g<topic>/{0}/\g<clientid>/\g<requestid>'

    @classmethod
    def request_topic(cls, topic, command, **fields):
        """Request topic from topic."""
        topic = os.path.join(topic, cls.REQUEST_SUFFIX)
        topic = common.topic_lazyformat(topic, command=command, **fields)
        return topic

    @classmethod
    def reply_topic(cls, topic, command, **fields):
        """Reply topic for topic."""
        topic = os.path.join(topic, cls.REPLY_SUFFIX)
        topic = common.topic_lazyformat(topic, command=command, **fields)
        return topic

    @classmethod
    def reply_topic_from_request(cls, request_topic):
        """Calculate request reply topic from request topic."""
        topic_repl = cls.TOPIC_REPL_FMT.format('reply')
        return cls.TOPIC_RE.sub(topic_repl, request_topic)

    @classmethod
    def request_topic_from_reply(cls, reply_topic):
        """Calculate request reply topic from request topic."""
        topic_repl = cls.TOPIC_REPL_FMT.format('request')
        return cls.TOPIC_RE.sub(topic_repl, reply_topic)

    @classmethod
    def clean_callback_fields(cls, fields_values):
        """Remove fields that should not be passed to callback."""
        # Remove 'request' fields
        cleaned_fields_values = {k: v for k, v in fields_values.items()
                                 if k not in cls.REQUEST_FIELDS}

        return cleaned_fields_values


class RequestServer(Topic):
    """Topic implementation for a Request server."""

    def __init__(self, topic, command, callback=None):
        assert callback
        request_topic = RequestTopic.request_topic(topic, command)
        super().__init__(request_topic, callback=callback)

    def wrap_callback(self, callback):  # overrides
        """Call topic callback with expanded field values.

        Publish return values to reply_topic.
        """
        @functools.wraps(callback)
        def _wrapper(mqttc, obj, msg):  # pylint:disable=unused-argument
            """Call topic callback with expanded field values.

            Publish return values to reply_topic.
            """
            # Remove 'request/client ids', not given to callback
            fields_values = self.fields_values(msg.topic)
            fields_values = RequestTopic.clean_callback_fields(fields_values)
            reply_topic = RequestTopic.reply_topic_from_request(msg.topic)

            # Add reply_publisher to message
            msg.reply_publisher = mqttc.publisher(reply_topic)

            result = callback(msg, **fields_values)

            # Asynchronous answer
            if result is None:
                return

            # Publish result
            msg.reply_publisher(result)

        return _wrapper


class RequestClient(Topic):
    """Topic implementation for a Request client."""

    REQUEST_TIMEOUT = 30
    REQUEST_ERRORS = {
        'timeout': 'Answer timeout',
        'pubtimeout': 'Request not published before timeout',
    }

    def __init__(self, topic, command, callback=None, clientid=None):
        assert callback is None
        assert clientid is not None

        self.clientid = clientid
        reply_topic = RequestTopic.reply_topic(topic, command,
                                               clientid=clientid)
        super().__init__(reply_topic, callback=callback)

        self.request_topic = RequestTopic.request_topic(topic, command,
                                                        clientid=clientid)
        self.requestid = 0
        self._request_lock = threading.Lock()
        self._request_event = threading.Event()
        self._request_answer = None

    def request(self, client, data, timeout=None, **fields):
        """Perform request under lock and wait for response."""
        with self._request_lock:  # pylint:disable=not-context-manager
            return self._request(client, data, timeout=timeout, **fields)

    def _request(self, client, data, timeout=None, **fields):
        """Perform request, should be called under lock.

        Create a new request by updating request_id.
        Register a callback for current request.
        Then perform blocking request."""
        self.requestid += 1
        topic = self.request_topic.format(requestid=self.requestid, **fields)
        reply_topic = RequestTopic.reply_topic_from_request(topic)

        with client.message_callback(reply_topic, self._cb_request):
            answer = self._do_blocking_request(client, topic, data, timeout)
            return answer

    def _do_blocking_request(self, client, topic, data, timeout=None):
        """Publish request and wait for request answer. Handles timeout."""
        timeout = timeout or self.REQUEST_TIMEOUT

        self._request_event.clear()
        self._request_answer = None

        message = client.publish(topic, data)

        answered = self._request_event.wait(timeout)
        if answered:
            return self._request_answer

        # No answer
        if message.is_published():
            raise RuntimeError(self.REQUEST_ERRORS['timeout'])
        else:
            raise RuntimeError(self.REQUEST_ERRORS['pubtimeout'])

    def _cb_request(self, mqttc, obj, msg):  # pylint:disable=W0221,W0613
        """Callback for request answer. Set answer and release event."""
        self._request_answer = msg.payload
        self._request_event.set()


class ChannelTopic(object):
    """ChannelTopic format."""
    OUTPUT_SUFFIX = 'data/out'
    INPUT_SUFFIX = 'data/in'

    INPUT_RE = re.compile('/%s$' % INPUT_SUFFIX)

    _TOPIC_RE = r'^{}/data/(in|out)'.format(r'(?P<topic>.*)')
    TOPIC_RE = re.compile(_TOPIC_RE)
    TOPIC_REPL = r'\g<topic>'

    @classmethod
    def output_topic(cls, topic):
        """Output topic from topic."""
        topic = os.path.join(topic, cls.OUTPUT_SUFFIX)
        return topic

    @classmethod
    def input_topic(cls, topic):
        """Input topic for topic."""
        topic = os.path.join(topic, cls.INPUT_SUFFIX)
        return topic

    @classmethod
    def output_topic_from_input(cls, input_topic):
        """Calculate output topic from input topic."""
        topic = cls.TOPIC_RE.sub(cls.TOPIC_REPL, input_topic)
        return cls.output_topic(topic)

    @classmethod
    def input_topic_from_output(cls, output_topic):
        """Calculate output topic from output topic."""
        topic = cls.TOPIC_RE.sub(cls.TOPIC_REPL, output_topic)
        return cls.input_topic(topic)


class OutputChannelServer(NullTopic):
    """Topic implementation for a Output only Channel server."""
    def __init__(self, topic, callback=None):
        super().__init__(topic, callback=callback)
        self.output_topic = ChannelTopic.output_topic(topic)

    def output_publisher(self, client, **fmt):
        """Output publisher function for topic formatted with **fmt."""
        topic = self.output_topic.format(**fmt)
        return client.publisher(topic)


class ChannelServer(Topic):
    """Topic implementation for a Channel server."""
    def __init__(self, topic, callback=None):
        input_topic = ChannelTopic.input_topic(topic)
        super().__init__(input_topic, callback=callback)
        self.output_topic = ChannelTopic.output_topic(topic)

    def output_publisher(self, client, **fmt):
        """Output publisher function for topic formatted with **fmt."""
        topic = self.output_topic.format(**fmt)
        return client.publisher(topic)


class ChannelClient(Topic):
    """Topic implementation for a Channel client."""
    def __init__(self, topic, callback=None):
        output_topic = ChannelTopic.output_topic(topic)
        super().__init__(output_topic, callback=callback)
        self.input_topic = ChannelTopic.input_topic(topic)

    def send(self, client, data, **fmt):
        """Send ``data`` to topic formatted with **fmt."""
        topic = self.input_topic.format(**fmt)
        return client.publish(topic, data)


class ErrorTopic(object):
    """ErrorTopic format."""
    ERROR_SUFFIX = 'error/'
    SUBSCRIBE_SUFFIX = 'error/#'

    @classmethod
    def error_topic(cls, topic):
        """Error base topic from topic."""
        topic = os.path.join(topic, cls.ERROR_SUFFIX)
        return topic

    @classmethod
    def subscribe_topic(cls, topic):
        """Error subscribe topic from topic."""
        topic = os.path.join(topic, cls.SUBSCRIBE_SUFFIX)
        return topic

    @staticmethod
    def relative_topic(base_topic, topic):
        """Get relative part from base_topic."""
        if topic.startswith(base_topic):
            return topic.replace(base_topic, '').lstrip('/')
        raise ValueError('Topic %s is not a subtopic from %s' %
                         (topic, base_topic))


class ErrorServer(NullTopic):
    """Topic implementation for an Error server."""

    def __init__(self, topic, callback=None):
        server_topic = ErrorTopic.error_topic(topic)
        super().__init__(server_topic, callback=callback)

        self.base_topic = topic

    def publish_error(self, client, topic, error_payload):
        """Publish ``error_payload`` to error topic for ``topic``."""
        relative_topic = ErrorTopic.relative_topic(self.base_topic, topic)
        error_topic = os.path.join(self.topic, relative_topic)
        client.publish(error_topic, error_payload)


class ErrorClient(Topic):
    """Topic implementation for an Error client."""

    def __init__(self, topic, callback=None):
        subscribe_topic = ErrorTopic.subscribe_topic(topic)
        super().__init__(subscribe_topic, callback=callback)

        self.error_topic = ErrorTopic.error_topic(topic)

    def wrap_callback(self, callback):  # overrides
        """Wrap callback to call it with relative topic."""
        @functools.wraps(callback)
        def _wrapper(mqttc, obj, msg):  # pylint:disable=unused-argument
            rel_topic = ErrorTopic.relative_topic(self.error_topic, msg.topic)
            return callback(msg, rel_topic)

        return _wrapper



# -*- coding: utf-8 -*-

"""IoT-LAB MQTT Node agent client"""

from __future__ import (absolute_import, division, print_function,
                        unicode_literals)
from builtins import *  # pylint:disable=W0401,W0614,W0622

import os

import iotlabmqtt.node
from iotlabmqtt import common
from iotlabmqtt import mqttcommon
from . import common as clientcommon

PARSER = common.MQTTAgentArgumentParser()
clientcommon.parser_add_site_arg(PARSER)


class NodeShell(clientcommon.CmdShell):
    """Node Agent Shell.

    :param client: mqttclient instance
    :param prefix: topics prefix
    :param site: agent site
    """

    RESET_USAGE = ('reset ARCHI NUM\n'
                   '  ARCHI: m3/a8\n'
                   '  NUM:   node num\n')

    UPDATE_USAGE = ('update ARCHI NUM FIRMWAREPATH\n'
                    '  ARCHI: m3/a8\n'
                    '  NUM:   node num\n'
                    '  FIRMWAREPATH: Path to the firmware.\n'
                    '                Idle firmware if not provided\n')

    POWERON_USAGE = ('poweron ARCHI NUM\n'
                     '  ARCHI: m3/a8\n'
                     '  NUM:   node num\n')
    POWEROFF_USAGE = ('poweroff ARCHI NUM\n'
                      '  ARCHI: m3/a8\n'
                      '  NUM:   node num\n')

    SERVER = iotlabmqtt.node.MQTTNodeAgent

    def __init__(self, client, prefix, site=None):
        assert site is not None
        super().__init__()

        clientid = clientcommon.clientid('nodeclient')

        staticfmt = {'site': site}
        _topics = mqttcommon.generate_topics_dict(
            self.SERVER.TOPICS, prefix, self.SERVER.AGENTTOPIC, staticfmt)

        _print_wrapper = self.async_print_handle_readlinebuff()
        error_cb = _print_wrapper(self.error_cb)

        self.topics = {
            'reset': mqttcommon.RequestClient(
                _topics['node'], 'reset', clientid=clientid),
            'update': mqttcommon.RequestClient(
                _topics['node'], 'update', clientid=clientid),
            'poweron': mqttcommon.RequestClient(
                _topics['node'], 'poweron', clientid=clientid),
            'poweroff': mqttcommon.RequestClient(
                _topics['node'], 'poweroff', clientid=clientid),

            'error': mqttcommon.ErrorClient(_topics['agenttopic'],
                                            callback=error_cb),
        }

        self.client = client
        self.client.topics = list(self.topics.values())

    def error_cb(self, message, relative_topic):  # pylint:disable=no-self-use
        """Callback on 'error' topic."""
        msg = message.payload.decode('utf-8')
        print('NODE ERROR: %s: %s' % (relative_topic, msg))

    @classmethod
    def from_opts_dict(cls, prefix, site, **kwargs):
        """Create class from argparse entries."""
        client = mqttcommon.MQTTClient.from_opts_dict(**kwargs)
        return cls(client, prefix=prefix, site=site)

    # # # # #
    # reset #
    # # # # #
    def do_reset(self, arg):
        """Reset node cpu: ARCHI NUM."""
        archi, num = self.cmd_split(arg)
        num = int(num)

        topic = self.topics['reset']
        ret = topic.request(self.client, b'', timeout=15,
                            archi=archi, num=num)
        if ret:
            raise RuntimeError(ret.decode('utf-8'))

    def help_reset(self):
        """Help reset command."""
        print(self.RESET_USAGE, end='')

    # # # # # #
    # Update  #
    # # # # # #
    def do_update(self, arg):
        """Update node firmware: ARCHI NUM FIRMWAREPATH."""
        splitted = self.cmd_split(arg)
        try:
            archi, num, firmware_path = splitted
        except ValueError:
            archi, num = splitted
            firmware = None  # Idle
        else:
            firmware = self._read_file(firmware_path, 'b')
        num = int(num)

        topic = self.topics['update']
        ret = topic.request(self.client, firmware, timeout=60,
                            archi=archi, num=num)
        if ret:
            raise RuntimeError(ret.decode('utf-8'))

    @staticmethod
    def _read_file(file_path, opt=''):
        """Open and read a file.

        Expands user directory
        """
        try:
            file_path = os.path.expanduser(file_path)  # expand '~'
            with open(file_path, 'r' + opt) as _fd:
                return _fd.read()
        except (IOError, OSError) as err:
            print('Could not open file: %s' % err)
            raise ValueError()

    def help_update(self):
        """Help update command."""
        print(self.UPDATE_USAGE, end='')

    # # # # # #
    # poweron #
    # # # # # #
    def do_poweron(self, arg):
        """Power ON node: ARCHI NUM."""
        archi, num = self.cmd_split(arg)
        num = int(num)

        topic = self.topics['poweron']
        ret = topic.request(self.client, b'', timeout=15,
                            archi=archi, num=num)
        if ret:
            raise RuntimeError(ret.decode('utf-8'))

    def help_poweron(self):
        """Help poweron command."""
        print(self.POWERON_USAGE, end='')

    # # # # # # #
    # poweroff  #
    # # # # # # #
    def do_poweroff(self, arg):
        """Power OFF node: ARCHI NUM."""
        archi, num = self.cmd_split(arg)
        num = int(num)

        topic = self.topics['poweroff']
        ret = topic.request(self.client, b'', timeout=15,
                            archi=archi, num=num)
        if ret:
            raise RuntimeError(ret.decode('utf-8'))

    def help_poweroff(self):
        """Help poweroff command."""
        print(self.POWEROFF_USAGE, end='')

    def run(self):
        """Run client and shell."""
        self.start()
        try:
            self.cmdloop()
        except KeyboardInterrupt:
            pass
        self.stop()

    def start(self):
        """Start Agent."""
        self.client.start()

    def stop(self):
        """Stop agent."""
        self.client.stop()


def main(opts=None):
    """SerialAgent shell client.

    :param opts: If provided, don't parse command line but use it instead
    :type opts: argparse Namespace object
    """
    opts = opts or PARSER.parse_args()
    shell = NodeShell.from_opts_dict(**vars(opts))
    shell.run()



# -*- coding: utf-8 -*-

"""IoT-LAB MQTT Log agent client"""

from __future__ import (absolute_import, division, print_function,
                        unicode_literals)
from builtins import *  # pylint:disable=W0401,W0614,W0622

import os
import base64

from iotlabmqtt import common
from iotlabmqtt import mqttcommon

from . import common as clientcommon

PARSER = common.MQTTAgentArgumentParser()


class LogShell(clientcommon.CmdShell):
    """Log messages Shell.

    :param client: mqttclient instance
    :param prefix: topics prefix
    """
    DEVNULL = open(os.devnull, 'w')

    OPEN_USAGE = ('open LOGFILE\n'
                  '  LOGFILE:   csv log file\n')
    CLOSE_USAGE = ('close\n'
                   '  Close current logfile\n')

    def __init__(self, client, prefix):
        super().__init__()

        _print_wrapper = self.async_print_handle_readlinebuff()
        log_cb = _print_wrapper(self.log_handler)

        self.topics = {
            'log': mqttcommon.LogTopic(prefix, log_cb),
        }

        self._logfile = self.DEVNULL

        self.client = client
        self.client.topics = list(self.topics.values())

    @property
    def logfile(self):
        """Current logfile."""
        return self._logfile

    @logfile.setter
    def logfile(self, log_file):
        """Set logfile and close previous one."""
        _logfile = self._logfile

        self._logfile = log_file

        if _logfile != self.DEVNULL:
            _logfile.close()

    @classmethod
    def from_opts_dict(cls, prefix, **kwargs):
        """Create class from argparse entries."""
        client = mqttcommon.MQTTClient.from_opts_dict(**kwargs)
        return cls(client, prefix=prefix)

    def do_open(self, arg):
        """Open a logfile."""
        try:
            logfile = open(arg, 'w', buffering=1)
        except (IOError, OSError) as err:
            print('Could not open file: %s' % err)
            raise ValueError()
        else:
            self.logfile = logfile

    def help_open(self):
        """Help open command."""
        print(self.OPEN_USAGE, end='')

    def do_close(self, _=None):
        """Close current logfile."""
        self.logfile = self.DEVNULL

    def help_close(self):
        """Help close command."""
        print(self.CLOSE_USAGE, end='')

    def log_handler(self, message):
        """Callback for logging."""
        self._log_print_message(message)
        self._log_save_message(message)

    @classmethod
    def _log_print_message(cls, message):
        """Print debug message in a readable format."""
        message_fmt = "%.3f %s '%s'"
        message_payload_str = cls._log_payload_str(message.payload)
        message_str = message_fmt % (message.timestamp, message.topic,
                                     message_payload_str)
        print(message_str)

    def _log_save_message(self, message):
        """Save message to logfile."""
        message_fmt = '%.3f;%s;%s\n'
        message_payload_str = base64.b64encode(message.payload)
        message_payload_str = message_payload_str.decode('utf-8')
        message_str = message_fmt % (message.timestamp, message.topic,
                                     message_payload_str)
        self.logfile.write(message_str)
        self.logfile.flush()

    @staticmethod
    def _log_payload_str(payload):
        """Payload representation."""
        try:
            payload_str = payload.decode('utf-8')
        except UnicodeDecodeError:
            payload_fmt = 'BIN: (len %u): hash: %s'
            short_sha1 = common.short_hash(payload)
            payload_str = payload_fmt % (len(payload), short_sha1)
        return payload_str

    def run(self):
        """Run client and shell."""
        self.start()
        try:
            self.cmdloop()
        except KeyboardInterrupt:
            pass
        self.stop()

    def start(self):
        """Start Agent."""
        self.client.start()

    def stop(self):
        """Stop agent."""
        self.client.stop()


def main(opts=None):
    """Logging shell client.

    :param opts: If provided, don't parse command line but use it instead
    :type opts: argparse Namespace object
    """
    opts = opts or PARSER.parse_args()
    shell = LogShell.from_opts_dict(**vars(opts))
    shell.run()



#! /usr/bin/env python
# -*- coding: utf-8 -*-

"""IoT-LAB MQTT Process client example
===================================

This example shows a process client. It runs two processes:

 * one that does echo stdin
 * one that print messages for 5 seconds

And then interracts with them.

.. note::

   This file is just an example that uses bare client library with arbitrary
   design choices and should not be taken as a good implementation reference.

Design choices
--------------

To simplify, only one subscribtion is done for each output topics.
And then, the ``procid`` part is handled by the generic callback.

The communication between the asynchronous callbacks and the main process is
done by using per process/topic Queues.

Running
-------

Example to run with a server with custom agenttopic:

::

   iotlab-mqtt-process --prefix PREFIX --agenttopic AGENTTOPIC BROKER

Then run the client with:

::

   python process_client.py --prefix PREFIX --agenttopic AGENTTOPIC BROKER


To see how topics are used, you can use the log client:

::

   iotlab-mqtt-clients log --prefix PREFIX BROKEr

"""

from __future__ import (absolute_import, division, print_function,
                        unicode_literals)
from builtins import *  # pylint:disable=W0401,W0614,W0622

import sys
import time
import json
import threading
import logging

import iotlabmqtt.process
from iotlabmqtt import common
from iotlabmqtt import mqttcommon
from iotlabmqtt.clients import common as clientcommon
from iotlabmqtt.processcommon import queue  # Python2/3 handled

# Add logging
logging.basicConfig(level=logging.DEBUG)

PARSER = common.MQTTAgentArgumentParser()
PARSER.add_argument('--agenttopic', help='Agent topic', required=True)

PROCESS_SERVER = iotlabmqtt.process.MQTTProcessAgent


class CustomProcessClient(object):
    """CustomProcessClient that runs two process and interracts with them.

    Asynchronous messages are stored in queues.
    Queues can be accessed with ``process_queue(procid, name)`` method.
    Where name is in ('stdout', 'stderr', 'retcode').
    """

    QUEUES = ('stdout', 'stderr', 'retcode')

    def __init__(self, client, prefix, agenttopic):
        _topics = mqttcommon.generate_topics_dict(PROCESS_SERVER.TOPICS,
                                                  prefix, agenttopic)

        clientid = clientcommon.clientid('customprocessclient')
        # Copied from clients/process.py
        self.topics = {
            'new': mqttcommon.RequestClient(
                _topics['agenttopic'], 'new', clientid=clientid),
            'list': mqttcommon.RequestClient(
                _topics['agenttopic'], 'list', clientid=clientid),
            'free': mqttcommon.RequestClient(
                _topics['agenttopic'], 'free', clientid=clientid),
            'freeall': mqttcommon.RequestClient(
                _topics['agenttopic'], 'freeall', clientid=clientid),

            # I use common callback for all processes
            'procrun': mqttcommon.RequestClient(
                _topics['process'], 'run', clientid=clientid),
            'procpoll': mqttcommon.RequestClient(
                _topics['process'], 'poll', clientid=clientid),
            'prockill': mqttcommon.RequestClient(
                _topics['process'], 'kill', clientid=clientid),
            'procrerun': mqttcommon.RequestClient(
                _topics['process'], 'rerun', clientid=clientid),

            'procstdin': mqttcommon.InputClient(_topics['procstdin']),
            'procstdout': mqttcommon.OutputClient(_topics['procstdout'],
                                                  callback=self._stdout_cb),
            'procstderr': mqttcommon.OutputClient(_topics['procstderr'],
                                                  callback=self._stderr_cb),

            'procret': mqttcommon.OutputClient(_topics['procret'],
                                               callback=self._returncode_cb),

            'error': mqttcommon.ErrorClient(_topics['agenttopic'],
                                            callback=self.error_cb),
        }
        self._rlock = threading.RLock()
        self._process_queues = {}

        self.client = client
        self.client.topics = list(self.topics.values())

    def error_cb(self, message, relative_topic):  # pylint:disable=no-self-use
        """Callback on 'error' topic."""
        msg = message.payload.decode('utf-8')
        print('PROCESS ERROR: %s: %s' % (relative_topic, msg), file=sys.stderr)

    def process_queue(self, procid, name):
        """Return given `procid` `name` process_queue.

        Create the Queue object if it does not currently exist.
        """
        assert name in self.QUEUES

        with self._rlock:
            proc_queues = self._process_queues.setdefault(procid, {})
            return proc_queues.setdefault(name, queue.Queue())

    def _save_payload(self, name, message, procid):
        """Store payload in 'self._process_queues[procid][name]' Queue."""
        logging.debug('_save_payload(%s, %s): %r', procid, name,
                      message.payload)
        proc_q = self.process_queue(procid, name)
        proc_q.put(message.payload)

    def _stdout_cb(self, message, procid):
        self._save_payload('stdout', message, procid)

    def _stderr_cb(self, message, procid):
        self._save_payload('stderr', message, procid)

    def _returncode_cb(self, message, procid):
        self._save_payload('retcode', message, procid)

    def start(self):
        """Start Agent."""
        self.client.start()

    def stop(self):
        """Stop agent."""
        self.client.stop()


ECHO_COMMAND = ['cat']
LONG_COMMAND = ['bash', '-c', "for i in {1..5}; do echo $i; sleep 1; done"]


def main(proc_client):
    try:
        process_client.start()
        _main(process_client)
    finally:
        process_client.stop()


def _main(proc_client):
    """Execute two processes and use their output.

    One first that echoes data, the other one that execute, returns
    an output and exit.

    :param proc_client: a started CustomProcessClient
    """
    # proc_client is already started

    echo_proc_name = 'echo'
    logging.info('Get a new process with name %s', echo_proc_name)
    ret = proc_client.topics['new'].request(
        proc_client.client, echo_proc_name.encode('utf-8'), timeout=5)
    ret_name = ret.decode('utf-8')
    assert echo_proc_name == ret_name, \
        'Proc new failed %s != %s' % (echo_proc_name, ret_name)
    logging.info('Success')

    logging.info('Get a new process without name')
    ret = proc_client.topics['new'].request(proc_client.client, b'', timeout=5)
    long_proc_name = ret.decode('utf-8')
    assert long_proc_name.isdigit(), 'Proc new failed %s' % long_proc_name
    logging.info('Success, name == %s', long_proc_name)

    logging.info('Start echo process: %s', ECHO_COMMAND)
    ret = proc_client.topics['procrun'].request(
        proc_client.client, json.dumps(ECHO_COMMAND).encode('utf-8'),
        procid=echo_proc_name, timeout=5)
    assert not ret

    ret = proc_client.topics['list'].request(proc_client.client, b'',
                                             timeout=5)
    logging.info('List of processes: %s', json.loads(ret.decode('utf-8')))

    logging.info('Start long process: %s', LONG_COMMAND)
    ret = proc_client.topics['procrun'].request(
        proc_client.client, json.dumps(LONG_COMMAND).encode('utf-8'),
        procid=long_proc_name, timeout=5)
    assert not ret

    # Interracting with long process
    logging.info('Wait long process stops')
    retcodequeue = process_client.process_queue(long_proc_name, 'retcode')
    stdoutqueue = process_client.process_queue(long_proc_name, 'stdout')

    ret = retcodequeue.get(True, 10)
    logging.info('Long process ret: %s', ret.decode('utf-8'))

    output = b''
    for i in range(0, 100):
        try:
            output += stdoutqueue.get_nowait()
            logging.debug('Get stdout Loop %s', i)
        except queue.Empty:
            break
    else:
        logging.info('Should have been out of loop before...')
    logging.info('Long process decoded output: %r', output.decode('utf-8'))

    # Interracting with echo process
    retcodequeue = process_client.process_queue(echo_proc_name, 'retcode')
    stdoutqueue = process_client.process_queue(echo_proc_name, 'stdout')
    for i in range(0, 3):
        logging.info('Sending message %s to echo stdin', i)
        message = ('Echo this: %s\n' % i).encode('utf-8')
        proc_client.topics['procstdin'].send(proc_client.client,
                                             message, procid=echo_proc_name)
        time.sleep(1)

    output = b''
    for i in range(0, 100):
        try:
            output += stdoutqueue.get_nowait()
            logging.debug('Get stdout Loop %s', i)
        except queue.Empty:
            break
    else:
        logging.info('Should have been out of loop before...')
    logging.info('Echo process decoded output: %r', output.decode('utf-8'))

    logging.info('Kill echo process')
    ret = proc_client.topics['prockill'].request(
        proc_client.client, b'', procid=echo_proc_name, timeout=5)
    assert not ret

    logging.info('Wait returncode')
    ret = retcodequeue.get(True, 10)
    logging.info('Echo process ret (killed): %s', ret.decode('utf-8'))

    logging.info('Free %s', long_proc_name)
    ret = proc_client.topics['free'].request(
        proc_client.client, long_proc_name.encode('utf-8'), timeout=5)
    assert not ret

    logging.info('Free %s', echo_proc_name)
    ret = proc_client.topics['free'].request(
        proc_client.client, echo_proc_name.encode('utf-8'), timeout=5)
    assert not ret

    ret = proc_client.topics['list'].request(proc_client.client, b'',
                                             timeout=5)
    logging.info('List of processes: %s', json.loads(ret.decode('utf-8')))


if __name__ == '__main__':
    opts = PARSER.parse_args()
    client = mqttcommon.MQTTClient.from_opts_dict(**vars(opts))
    process_client = CustomProcessClient(client, prefix=opts.prefix,
                                         agenttopic=opts.agenttopic)
    main(process_client)



# -*- coding: utf-8 -*-

"""IoT-LAB MQTT Process agent client"""

from __future__ import (absolute_import, division, print_function,
                        unicode_literals)
from builtins import *  # pylint:disable=W0401,W0614,W0622

import json

import iotlabmqtt.process
from iotlabmqtt import common
from iotlabmqtt import mqttcommon

from . import common as clientcommon


PARSER = common.MQTTAgentArgumentParser()
clientcommon.parser_add_site_or_agenttopic_arg(PARSER)


class ProcessShell(clientcommon.CmdShell):
    # pylint:disable=too-many-public-methods
    """Process Agent Shell.

    :param client: mqttclient instance
    :param prefix: topics prefix
    :param site: agent site
    """
    NEW_USAGE = ('new [NAME]\n'
                 '  NAME: optional process identifier\n')
    FREE_USAGE = ('free NAME\n'
                  '  NAME: process identifier\n')
    LIST_USAGE = ('list\n')

    RUN_USAGE = ('run NAME CMD0 [CMD1 [...]]\n'
                 '  NAME: process identifier\n'
                 '  CMD: command arguments\n')
    POLL_USAGE = ('poll NAME\n'
                  '  NAME: process identifier\n')
    KILL_USAGE = ('kill NAME\n'
                  '  NAME: process identifier\n')
    RERUN_USAGE = ('rerun NAME\n'
                   '  NAME: process identifier\n')

    STDINLINE_USAGE = (
        'stdinline NAME message with spaces\n'
        '  NAME: process identifier\n'
        '  Message is sent as utf-8 data with a \\n character\n'
    )
    STDINRAW_USAGE = (
        'stdinraw NAME message with spaces\n'
        '  NAME: process identifier\n'
        '  Message is sent as utf-8 data\n'
    )

    SERVER = iotlabmqtt.process.MQTTProcessAgent

    def __init__(self, client, prefix, site=None, agenttopic=None):
        assert site is not None or agenttopic is not None
        super().__init__()

        clientid = clientcommon.clientid('processclient')

        _topics = self._topics_dict(prefix, site, agenttopic)

        _print_wrapper = self.async_print_handle_readlinebuff()
        stdout_cb = _print_wrapper(self.stdout_cb)
        stderr_cb = _print_wrapper(self.stderr_cb)
        returncode_cb = _print_wrapper(self.returncode_cb)
        error_cb = _print_wrapper(self.error_cb)

        self.topics = {
            'new': mqttcommon.RequestClient(
                _topics['agenttopic'], 'new', clientid=clientid),
            'list': mqttcommon.RequestClient(
                _topics['agenttopic'], 'list', clientid=clientid),
            'free': mqttcommon.RequestClient(
                _topics['agenttopic'], 'free', clientid=clientid),
            'freeall': mqttcommon.RequestClient(
                _topics['agenttopic'], 'freeall', clientid=clientid),

            'procrun': mqttcommon.RequestClient(
                _topics['process'], 'run', clientid=clientid),
            'procpoll': mqttcommon.RequestClient(
                _topics['process'], 'poll', clientid=clientid),
            'prockill': mqttcommon.RequestClient(
                _topics['process'], 'kill', clientid=clientid),
            'procrerun': mqttcommon.RequestClient(
                _topics['process'], 'rerun', clientid=clientid),

            'procstdin': mqttcommon.InputClient(_topics['procstdin']),
            'procstdout': mqttcommon.OutputClient(_topics['procstdout'],
                                                  callback=stdout_cb),
            'procstderr': mqttcommon.OutputClient(_topics['procstderr'],
                                                  callback=stderr_cb),

            'procret': mqttcommon.OutputClient(_topics['procret'],
                                               callback=returncode_cb),

            'error': mqttcommon.ErrorClient(_topics['agenttopic'],
                                            callback=error_cb),
        }

        self.client = client
        self.client.topics = list(self.topics.values())

    def _topics_dict(self, prefix, site=None, agenttopic=None):
        """Return topics formated for site or agenttopic.

        Allows overwriting default prefix with agenttopic.
        """

        # Prefix topics with 'agenttopic'
        agenttopic = (agenttopic if agenttopic is not None else
                      self.SERVER.AGENTTOPIC)
        staticfmt = {'site': site} if site is not None else {}

        err = 'site required for agenttopic format'
        assert not ('site' in agenttopic and site is None), err

        return mqttcommon.generate_topics_dict(
            self.SERVER.TOPICS, prefix, agenttopic, staticfmt)

    def error_cb(self, message, relative_topic):  # pylint:disable=no-self-use
        """Callback on 'error' topic."""
        msg = message.payload.decode('utf-8')
        print('PROCESS ERROR: %s: %s' % (relative_topic, msg))

    @classmethod
    def from_opts_dict(cls, prefix, site=None, agenttopic=None, **kwargs):
        """Create class from argparse entries."""
        client = mqttcommon.MQTTClient.from_opts_dict(**kwargs)
        return cls(client, prefix=prefix, site=site, agenttopic=agenttopic)

    def stdout_cb(self, message, procid):
        """Handle messages received from stdout."""
        self._fd_cb('stdout', procid, message)

    def stderr_cb(self, message, procid):
        """Handle messages received from stderr."""
        self._fd_cb('stderr', procid, message)

    def _fd_cb(self, fdname, procid, message):
        """Print `message` payload for `fdname` and `procid`."""
        data = message.payload.decode('utf-8', 'replace')
        data = self._repr_without_u_for_unicode_strings(data)
        print('%s(%s): %s' % (fdname, procid, data))

    @staticmethod
    def _repr_without_u_for_unicode_strings(unicode_message):
        """Returns string 'repr' without the leading 'u' for unicode.

        If python2, repr of an unicode string starts with a 'u', remove it
        """
        message = repr(unicode_message)
        if message.startswith('u'):
            message = message[1:]
        return message

    def returncode_cb(self, message, procid):  # pylint:disable=no-self-use
        """Handle returncode sent by process."""
        returncode = message.payload.decode('utf-8', 'replace')
        print('closed(%s): %s' % (procid, returncode))

    def do_new(self, arg):
        """Allocate a new process: [NAME]."""
        args = self.cmd_split(arg)
        if args:
            (name,) = args
        else:
            name = ''

        payload = name.encode('utf-8')  # can be empty

        topic = self.topics['new']
        ret = topic.request(self.client, payload, timeout=5)

        print(ret.decode('utf-8'))

    def help_new(self):
        """Help new command."""
        print(self.NEW_USAGE, end='')

    def do_free(self, arg):
        """Free given process: NAME."""
        (name,) = self.cmd_split(arg)

        payload = name.encode('utf-8')  # can be empty

        topic = self.topics['free']
        ret = topic.request(self.client, payload, timeout=5)
        if ret:
            raise RuntimeError(ret.decode('utf-8'))

    def help_free(self):
        """Help free command."""
        print(self.FREE_USAGE, end='')

    def do_list(self, _):
        """List processes names."""
        topic = self.topics['list']
        ret = topic.request(self.client, b'', timeout=5)

        # May have errors here but I don't know how to generate it
        # Ignore, its just a simple client
        names = json.loads(ret.decode('utf-8'))
        print([str(n) for n in names])

    def help_list(self):
        """Help list command."""
        print(self.LIST_USAGE, end='')

    def do_run(self, arg):
        """Run command for process name: NAME COMMAND."""
        args = self.cmd_split(arg)
        try:
            name = args[0]
            command = [args[1]] + args[2:]
        except IndexError:
            raise ValueError

        payload = json.dumps(command).encode('utf-8')

        topic = self.topics['procrun']
        ret = topic.request(self.client, payload, timeout=10, procid=name)
        if ret:
            raise RuntimeError(ret.decode('utf-8'))

    def help_run(self):
        """Help run command."""
        print(self.RUN_USAGE, end='')

    def do_poll(self, arg):
        """Poll process name: NAME."""
        (name,) = self.cmd_split(arg)

        topic = self.topics['procpoll']
        ret = topic.request(self.client, b'', timeout=10, procid=name)
        if ret:
            print(ret.decode('utf-8'))

    def help_poll(self):
        """Help poll command."""
        print(self.POLL_USAGE, end='')

    def do_kill(self, arg):
        """Kill process name: NAME."""
        (name,) = self.cmd_split(arg)

        topic = self.topics['prockill']
        ret = topic.request(self.client, b'', timeout=10, procid=name)
        if ret:
            raise RuntimeError(ret.decode('utf-8'))

    def help_kill(self):
        """Help kill command."""
        print(self.KILL_USAGE, end='')

    def do_rerun(self, arg):
        """Rerun process name: NAME."""
        (name,) = self.cmd_split(arg)

        topic = self.topics['procrerun']
        ret = topic.request(self.client, b'', timeout=10, procid=name)
        if ret:
            raise RuntimeError(ret.decode('utf-8'))

    def help_rerun(self):
        """Help rerun command."""
        print(self.RERUN_USAGE, end='')

    def do_stdinline(self, arg):
        """Send message to process `name` stdin: NAME MESSAGE.

        Add a newline character
        """
        name, message = self.cmd_split(arg, 1)

        payload = (message + '\n').encode('utf-8')

        self.topics['procstdin'].send(self.client, payload, procid=name)

    def help_stdinline(self):
        """Help stdinline command."""
        print(self.STDINLINE_USAGE, end='')

    def do_stdinraw(self, arg):
        """Send message to process `name` stdin: NAME MESSAGE."""
        name, message = self.cmd_split(arg, 1)

        payload = (message).encode('utf-8')

        self.topics['procstdin'].send(self.client, payload, procid=name)

    def help_stdinraw(self):
        """Help stdinraw command."""
        print(self.STDINRAW_USAGE, end='')

    def run(self):
        """Run client and shell."""
        self.start()
        try:
            self.cmdloop()
        except KeyboardInterrupt:
            pass
        self.stop()

    def start(self):
        """Start Agent."""
        self.client.start()

    def stop(self):
        """Stop agent."""
        self.client.stop()


def main(opts=None):
    """ProcessAgent shell client.

    :param opts: If provided, don't parse command line but use it instead
    :type opts: argparse Namespace object
    """
    opts = opts or PARSER.parse_args()
    clientcommon.parser_verify_site_and_agenttopic(PARSER, opts)
    shell = ProcessShell.from_opts_dict(**vars(opts))
    shell.run()



# -*- coding: utf-8 -*-

r"""IoT-LAB MQTT process agent
=============================

Process agent allows running and managing processes.

Process agent base topic: ::

   {prefix}/iot-lab/process/{site}

Every topics from process agent topics start by this topic prefix

:param prefix: configured prefix
:param site: site where the agent is run

.. note::

   ``iot-lab/process/{site}`` part can be replaced with the ``--agenttopic``
   option to allow running it on other servers than IoT-LAB.

Topics Summary
==============

.. |proc|             replace::  ``{processagenttopic}``

.. |ctlnew|           replace::  |proc|\ ``/ctl/new``
.. |ctllist|          replace::  |proc|\ ``/ctl/list``
.. |ctlfree|          replace::  |proc|\ ``/ctl/free``

.. |procid|           replace::  |proc|\ ``/process/{procid}``
.. |procctlrun|       replace::  |procid|\ ``/ctl/run``
.. |procctlpoll|      replace::  |procid|\ ``/ctl/poll``
.. |procctlkill|      replace::  |procid|\ ``/ctl/kill``
.. |procctlrerun|     replace::  |procid|\ ``/ctl/rerun``

.. |fdin|             replace::  ``fd/stdin``
.. |fdout|            replace::  ``fd/stdout``
.. |fderr|            replace::  ``fd/stderr``
.. |procstdin|        replace::  |procid|\ /\ |fdin|
.. |procstdout|       replace::  |procid|\ /\ |fdout|
.. |procstderr|       replace::  |procid|\ /\ |fderr|

.. |retcode|          replace::  ``event/returncode``
.. |procret|          replace::  |procid|\ /\ |retcode|

.. |error_t|          replace::  |proc|\ ``/error/``
.. |request|          replace::  :ref:`Request <RequestTopic>`
.. |error|            replace::  :ref:`Error <ErrorTopic>`

.. |request_topic|    replace::  *{topic}*/**request/{clientid}/{requestid}**
.. |reply_topic|      replace::  *{topic}*/**reply/{clientid}/{requestid}**

.. |outtopic|         replace::  Output
.. |intopic|          replace::  Input

+-+---------------------------------------------------------------+-----------+
| |Topic                                                          | Type      |
+=+===============================================================+===========+
|  **Process agent**                                                          |
+-+---------------------------------------------------------------+-----------+
| ``{prefix}/iot-lab/process/{site}``                                         |
+-+---------------------------------------------------------------+-----------+
| ||error_t|                                                      | |error|   |
+-+---------------------------------------------------------------+-----------+
|  **Process ids management**                                                 |
+-+---------------------------------------------------------------+-----------+
| ||ctlnew|                                                       ||request|  |
+-+---------------------------------------------------------------+-----------+
| ||ctlfree|                                                      ||request|  |
+-+---------------------------------------------------------------+-----------+
| ||ctllist|                                                      ||request|  |
+-+---------------------------------------------------------------+-----------+
|  **Process control**                                                        |
+-+---------------------------------------------------------------+-----------+
| ||procctlrun|                                                   ||request|  |
+-+---------------------------------------------------------------+-----------+
| ||procctlpoll|                                                  ||request|  |
+-+---------------------------------------------------------------+-----------+
| ||procctlkill|                                                  ||request|  |
+-+---------------------------------------------------------------+-----------+
| ||procctlrerun|                                                 ||request|  |
+-+---------------------------------------------------------------+-----------+
|  **Process file descriptors**                                               |
+-+---------------------------------------------------------------+-----------+
| ||procstdin|                                                    | |intopic| |
+-+---------------------------------------------------------------+-----------+
| ||procstdout|                                                   | |outtopic||
+-+---------------------------------------------------------------+-----------+
| ||procstderr|                                                   | |outtopic||
+-+---------------------------------------------------------------+-----------+
|  **Process events**                                                         |
+-+---------------------------------------------------------------+-----------+
| ||procret|                                                      | |outtopic||
+-+---------------------------------------------------------------+-----------+


Process Agent global topics
===========================

Error Topic
-----------

Asynchronous error messages are posted to error topic.
Failure on requests are not duplicated here.

* |error_t|


For format see: :ref:`ErrorTopic`


Process ids management topics
=============================

To run a process, a process id should first be allocated.
This id will be used for every commands on the process.

.. note::

   Seperating this from the ``procrun`` command allows registering first to
   all interesting topics (e.g. ``fd/stdout``, ``event/returncode``) so no
   messages are missed.


New process id
--------------

Allocate a new process id, if a ``procid`` is given it will use this one or
if empty, an uniq integer one will be provided.

Process ids will either be the provided ones, or a string representation of
integers.

:param procid: ``procid`` can be given or empty to get an uniq ``int`` one.
               An error is returned if given one is already used.
               It cannot contain ``{}/+#``.
:type procid:  ``utf-8`` encoded string
:returns:      ``utf-8`` allocated procid, which is either the given one or
               an ``int`` string.
               If message does not match above rules, its an error message.


+-----------------------------------------------------------------------------+
| ``new`` request:                                                            |
+============+================================================================+
| Topic:     |    |ctlnew|                                                    |
+------------+----------------------------------------+-----------------------+
|**Message** | **Topic**                              | **Payload**           |
+------------+----------------------------------------+-----------------------+
| Request    | |request_topic|                        | *empty* or ``procid`` |
+------------+----------------------------------------+-----------------------+
| Reply      | |reply_topic|                          |``procid`` or error_msg|
+------------+----------------------------------------+-----------------------+


Free process id
---------------

Free given process id. Process should be idle.

:param procid: ``utf-8`` ``procid`` to release
:type procid: ``utf-8`` encoded string


+-----------------------------------------------------------------------------+
| ``free`` request:                                                           |
+============+================================================================+
| Topic:     |    |ctlfree|                                                   |
+------------+-----------------------------------------+----------------------+
|**Message** | **Topic**                               | **Payload**          |
+------------+-----------------------------------------+----------------------+
| Request    | |request_topic|                         | ``procid``           |
+------------+-----------------------------------------+----------------------+
| Reply      | |reply_topic|                           | *empty* or error_msg |
+------------+-----------------------------------------+----------------------+


List processes ids
------------------

List processes ids.

:returns: ``utf-8`` encoded json list of ids

+-----------------------------------------------------------------------------+
| ``list`` request:                                                           |
+============+================================================================+
| Topic:     |    |ctllist|                                                   |
+------------+-----------------------------------------+----------------------+
|**Message** | **Topic**                               | **Payload**          |
+------------+-----------------------------------------+----------------------+
| Request    | |request_topic|                         | *empty*              |
+------------+-----------------------------------------+----------------------+
| Reply      | |reply_topic|                           | ``utf-8 json list``  |
+------------+-----------------------------------------+----------------------+


Process control
===============

Control process execution.

Process run
-----------

Run given command for given process id. The process should currently be idle.

:param command: ``utf-8`` encoded json list of arguments

.. warning::
   Care should be taken when running an *infinite running* process with
   **sudo** as it cannot be killed by its original user. `Read more`_

.. _Read more: https://www.reddit.com/r/linux/comments/2av912/

+-----------------------------------------------------------------------------+
| ``run`` request:                                                            |
+============+================================================================+
| Topic:     |    |procctlrun|                                                |
+------------+-----------------------------------------+----------------------+
|**Message** | **Topic**                               | **Payload**          |
+------------+-----------------------------------------+----------------------+
| Request    | |request_topic|                         | *command*            |
+------------+-----------------------------------------+----------------------+
| Reply      | |reply_topic|                           | *empty* or error_msg |
+------------+-----------------------------------------+----------------------+


Process poll
------------

Check if given process is still running.
If stopped, it returns returncode else an empty message.

:returns: empty if running, returncode ``int`` string if stopped or
          an error message.

+-----------------------------------------------------------------------------+
| ``poll`` request:                                                           |
+============+================================================================+
| Topic:     |    |procctlpoll|                                               |
+------------+-----------------------------------------+----------------------+
|**Message** | **Topic**                               | **Payload**          |
+------------+-----------------------------------------+----------------------+
| Request    | |request_topic|                         | *empty*              |
+------------+-----------------------------------------+----------------------+
| Reply      | |reply_topic|                           | *empty* or           |
|            |                                         | ``returncode`` or    |
|            |                                         | error_msg            |
+------------+-----------------------------------------+----------------------+


Process kill
------------

Kill given process.

.. note::
   A SIGTERM is first sent, then a SIGKILL one second after if still running.

+-----------------------------------------------------------------------------+
| ``kill`` request:                                                           |
+============+================================================================+
| Topic:     |    |procctlkill|                                               |
+------------+-----------------------------------------+----------------------+
|**Message** | **Topic**                               | **Payload**          |
+------------+-----------------------------------------+----------------------+
| Request    | |request_topic|                         | *empty*              |
+------------+-----------------------------------------+----------------------+
| Reply      | |reply_topic|                           | *empty* or error_msg |
+------------+-----------------------------------------+----------------------+


Process rerun
-------------

Re-run given process with the same command.
The process should currently be idle.

+-----------------------------------------------------------------------------+
| ``rerun`` request:                                                          |
+============+================================================================+
| Topic:     |    |procctlrerun|                                              |
+------------+-----------------------------------------+----------------------+
|**Message** | **Topic**                               | **Payload**          |
+------------+-----------------------------------------+----------------------+
| Request    | |request_topic|                         | *empty*              |
+------------+-----------------------------------------+----------------------+
| Reply      | |reply_topic|                           | *empty* or error_msg |
+------------+-----------------------------------------+----------------------+


Process file descriptors
========================

Process standard input, output and error topics.
They transport raw data to allow exchanging everything with the application.

This implies that messages may not be ``utf-8`` data,
or not seperated on new lines characters or even split in the middle of
an ``utf-8`` code-point.

Sent messages are only empty on process termination,
where ``fd/stdout`` and ``fd/stderr`` receive a final empty message.
During normal execution output messages are never empty.


+-----------------------------------------------------------------------------+
| **Process file descriptors**                                                |
+============+================================================================+
| Topic:     |    |procid|                                                    |
+------------+-----------------------------------------+----------------------+
|**Message** | **Topic**                               | **Payload**          |
+------------+-----------------------------------------+----------------------+
| Input      | |fdin|                                  | RAW bytes            |
+------------+-----------------------------------------+----------------------+
| Output     | |fdout|                                 | RAW bytes            |
+------------+-----------------------------------------+----------------------+
| Output     | |fderr|                                 | RAW bytes            |
+------------+-----------------------------------------+----------------------+


Process events
==============

Sends process returncode as a ``int`` string on process termination.

+-----------------------------------------------------------------------------+
| **Events  file descriptors**                                                |
+============+================================================================+
| Topic:     |    |procid|                                                    |
+------------+-----------------------------------------+----------------------+
|**Message** | **Topic**                               | **Payload**          |
+------------+-----------------------------------------+----------------------+
| Output     | |retcode|                               | ``returncode string``|
+------------+-----------------------------------------+----------------------+
"""

from __future__ import (absolute_import, division, print_function,
                        unicode_literals)
from builtins import *  # pylint:disable=W0401,W0614,W0622

import json

from . import common
from . import mqttcommon
from . import processcommon

PARSER = common.MQTTAgentArgumentParser()
PARSER.add_agenttopic_argument()


class MQTTProcessAgent(object):
    """Process Manager Agent implementation for MQTT."""
    AGENTTOPIC = 'iot-lab/process/{site}'
    TOPICS = {
        'process': 'process/{procid}',
        'procstdin': 'process/{procid}/fd/stdin',
        'procstdout': 'process/{procid}/fd/stdout',
        'procstderr': 'process/{procid}/fd/stderr',
        'procret': 'process/{procid}/event/returncode',
    }
    HOSTNAME = common.hostname()

    def __init__(self, client, prefix='', agenttopic=None):
        super().__init__()

        staticfmt = {'site': self.HOSTNAME}
        agenttopic = agenttopic if agenttopic is not None else self.AGENTTOPIC
        _topics = mqttcommon.generate_topics_dict(self.TOPICS, prefix,
                                                  agenttopic, staticfmt)

        self.topics = {
            'new': mqttcommon.RequestServer(
                _topics['agenttopic'], 'new', callback=self.cb_new),
            'free': mqttcommon.RequestServer(
                _topics['agenttopic'], 'free', callback=self.cb_free),
            'list': mqttcommon.RequestServer(
                _topics['agenttopic'], 'list', callback=self.cb_list),

            # Used for errors
            'process': mqttcommon.NullTopic(_topics['process']),
            'procrun': mqttcommon.RequestServer(
                _topics['process'], 'run', callback=self.cb_procrun),
            'procpoll': mqttcommon.RequestServer(
                _topics['process'], 'poll', callback=self.cb_procpoll),
            'prockill': mqttcommon.RequestServer(
                _topics['process'], 'kill', callback=self.cb_prockill),
            'procrerun': mqttcommon.RequestServer(
                _topics['process'], 'rerun', callback=self.cb_procrerun),

            'procstdin': mqttcommon.InputServer(_topics['procstdin'],
                                                callback=self.cb_procstdin),
            'procstdout': mqttcommon.OutputServer(_topics['procstdout']),
            'procstderr': mqttcommon.OutputServer(_topics['procstderr']),
            'procret': mqttcommon.OutputServer(_topics['procret']),

            'error': mqttcommon.ErrorServer(_topics['agenttopic']),
        }

        proc_callbacks = (self._closed_cb, self._stdout_cb, self._stderr_cb,
                          self._error_cb)
        self.process = processcommon.ProcessManager(callbacks=proc_callbacks)

        self.client = client
        self.client.topics = list(self.topics.values())

    def error(self, topic, message):
        """Publish error that happend on topic."""
        self.topics['error'].publish_error(self.client, topic,
                                           message.encode('utf-8'))

    def cb_new(self, message):
        """Alloc a new process id.

        If a payload is given, try to use it as processid name.
        Name cannot contain '/{}+#' characters.

        :returns: process id
        """
        try:
            name = message.payload.decode('utf-8')
            self._valid_name(name)
            return self.process.new(name).encode('utf-8')
        except UnicodeError:
            return 'Error decoding utf-8 payload'.encode('utf-8')
        except (KeyError, ValueError) as err:
            return err.args[0].encode('utf-8')

    @staticmethod
    def _valid_name(name=''):
        """Check 'name' is valid == no MQTT special characters."""
        invalid_chars = '/{}+#'
        valid_name = set(invalid_chars).isdisjoint(set(name))
        if not valid_name:
            raise ValueError('Name should not contain %s' % invalid_chars)

    def cb_free(self, message):
        """Free given process id.

        Process should be idle.
        """
        try:
            name = message.payload.decode('utf-8')
            self.process.free(name)
            return ''.encode('utf-8')
        except UnicodeError:
            return 'Error decoding utf-8 payload'.encode('utf-8')
        except KeyError:
            return ('Name %s not found' % name).encode('utf-8')
        except RuntimeError as err:
            return err.args[0].encode('utf-8')

    def cb_list(self, _):
        """List current processes."""
        return json.dumps(self.process.list()).encode('utf-8')

    # Process commands

    def cb_procrun(self, message, procid):
        """Run command for process id.

        Message payload should be the json encoded list of arguments.
        """
        try:
            proc = self.process[procid]
            command = json.loads(message.payload.decode('utf-8'))
            return proc.run(command).encode('utf-8')
        except UnicodeError:
            return 'Error decoding utf-8 payload'.encode('utf-8')
        except (ValueError, TypeError):
            err = 'Invalid command, should be a json encoded list'
            return err.encode('utf-8')
        except KeyError:
            return ('Name %s not found' % procid).encode('utf-8')

    def cb_procpoll(self, _, procid):
        """Check if given process is finished.

        :returns: process returncode if finished.
        """
        try:
            proc = self.process[procid]
            return proc.poll().encode('utf-8')
        except KeyError:
            return ('Name %s not found' % procid).encode('utf-8')

    def cb_prockill(self, _, procid):
        """Kill given process."""
        try:
            proc = self.process[procid]
            return proc.kill().encode('utf-8')
        except KeyError:
            return ('Name %s not found' % procid).encode('utf-8')

    def cb_procrerun(self, _, procid):
        """Rerun given process with the same arguments."""
        try:
            proc = self.process[procid]
            return proc.rerun().encode('utf-8')
        except KeyError:
            return ('Name %s not found' % procid).encode('utf-8')

    def cb_procstdin(self, message, procid):
        """Send message to given process standard input."""
        try:
            proc = self.process[procid]
            proc.write(message.payload)
        except KeyError:
            topic = self.topics['procstdin'].topic.format(procid=procid)
            self.error(topic, 'No process on %s' % procid)

    # Process callbacks

    def _stdout_cb(self, name, data):
        """Process callback for stdout data."""
        self.topics['procstdout'].send(self.client, data, procid=name)

    def _stderr_cb(self, name, data):
        """Process callback for stderr data."""
        self.topics['procstderr'].send(self.client, data, procid=name)

    def _closed_cb(self, name, returncode):
        """Process callback when process closes."""
        data = ('%s' % returncode).encode('utf-8')
        self.topics['procret'].send(self.client, data, procid=name)

    def _error_cb(self, name, message):
        """Process callback when an error occurs."""
        topic = self.topics['process'].topic.format(procid=name)
        self.error(topic, message)

    # Agent running

    def run(self):
        """Run agent."""
        try:
            self.start()
            common.wait_sigint()
        finally:
            self.stop()

    def start(self):
        """Start Agent."""
        self.client.start()
        self.process.start()

    def stop(self):
        """Stop agent."""
        self.process.stop()
        self.client.stop()

    @classmethod
    def from_opts_dict(cls, prefix, agenttopic=None, **kwargs):
        """Create class from argparse entries."""
        client = mqttcommon.MQTTClient.from_opts_dict(**kwargs)
        return cls(client, prefix, agenttopic=agenttopic)


def main():
    """Run process agent."""
    opts = PARSER.parse_args()
    aggr = MQTTProcessAgent.from_opts_dict(**vars(opts))
    aggr.run()



# -*- coding:utf-8 -*-

"""MQTTCommon module tests."""

from __future__ import (absolute_import, division, print_function,
                        unicode_literals)
from builtins import *  # pylint:disable=W0401,W0614,W0622

from iotlabmqtt.clients import common as clientcommon
from . import TestCaseImproved


class TestCmdShell(TestCaseImproved):
    """Test CmdShell methods."""

    def test_cmd_split(self):
        """Test cmd_split method."""
        cmd_split = clientcommon.CmdShell.cmd_split

        # One command
        args = cmd_split('command')
        self.assertEqual(args, ['command'])
        args = cmd_split('command   ')
        self.assertEqual(args, ['command'])

        # Simple arguments
        args = cmd_split('command arg1 arg2 arg3')
        self.assertEqual(args, ['command', 'arg1', 'arg2', 'arg3'])
        args = cmd_split('command arg1 arg2 arg3', maxsplit=1)
        self.assertEqual(args, ['command', 'arg1 arg2 arg3'])

        # With more whitespaces
        args = cmd_split('command   arg1    arg2    arg3')
        self.assertEqual(args, ['command', 'arg1', 'arg2', 'arg3'])
        args = cmd_split('command   arg1    arg2    arg3', maxsplit=1)
        self.assertEqual(args, ['command', 'arg1    arg2    arg3'])
        args = cmd_split('command   arg1    arg2    arg3', maxsplit=2)
        self.assertEqual(args, ['command', 'arg1', 'arg2    arg3'])

        # Empty commands
        args = cmd_split('')
        self.assertEqual(args, [])
        args = cmd_split('   ')
        self.assertEqual(args, [])



# -*- coding:utf-8 -*-

"""IoT-LAB MQTT Radio Sniffer agent client"""

from __future__ import (absolute_import, division, print_function,
                        unicode_literals)
from builtins import *  # pylint:disable=W0401,W0614,W0622

import threading

import iotlabmqtt.radiosniffer
from iotlabmqtt import common
from iotlabmqtt import mqttcommon

from . import common as clientcommon

PARSER = common.MQTTAgentArgumentParser()
clientcommon.parser_add_site_arg(PARSER)


class PcapFiles(object):
    """Handle pcap files by type.

    Wraps a dict of pcap files with an interface to allow correct management
    and data integrity. It is Thread safe.
    """
    def __init__(self):
        super().__init__()
        self.files = {}
        self._lock = threading.Lock()

    def clear(self, pcaptype):
        """Remove given file and close it."""
        with self._lock:  # pylint:disable=not-context-manager
            return self._clear(pcaptype)

    def _clear(self, pcaptype):
        try:
            self.files[pcaptype].close()
        except KeyError:
            pass
        finally:
            self.files.pop(pcaptype, None)

    def set(self, pcaptype, pcapfd, pcap_header):
        """Set pcap file as `pcaptype` and write its header."""
        with self._lock:  # pylint:disable=not-context-manager
            return self._set(pcaptype, pcapfd, pcap_header)

    def _set(self, pcaptype, pcapfd, pcap_header):
        self._clear(pcaptype)

        self.files[pcaptype] = pcapfd

        self._write(pcaptype, pcap_header)

    def write(self, pcaptype, data):
        """Write to given `pcaptype` file if its set."""
        with self._lock:  # pylint:disable=not-context-manager
            self._write(pcaptype, data)

    def _write(self, pcaptype, data):
        try:
            self.files[pcaptype].write(data)
            self.files[pcaptype].flush()
        except KeyError:
            pass


class RadioSnifferShell(clientcommon.CmdShell):
    """Radio Sniffer Agent Shell.

    :param host: broker host
    :param port: broker port
    :param prefix: topics prefix
    :param site: agent site
    """

    RAWSTART_USAGE = ('rawstart ARCHI NUM CHANNEL\n'
                      '  ARCHI:   m3/a8\n'
                      '  NUM:     node num\n'
                      '  CHANNEL: sniffer channel\n')
    RAWPCAP_USAGE = ('rawpcap FILEPATH\n'
                     '  FILEPATH: rawpcap output\n')
    RAWPCAPCLOSE_USAGE = ('rawpcapclose\n'
                          '  Close the current rawpcap file\n')

    STOP_USAGE = ('stop ARCHI NUM\n'
                  '  ARCHI: m3/a8\n'
                  '  NUM:   node num\n')

    STOPALL_USAGE = 'stopall\n'

    SERVER = iotlabmqtt.radiosniffer.MQTTRadioSnifferAggregator

    def __init__(self, client, prefix='', site=None):
        assert site is not None
        super().__init__()

        clientid = clientcommon.clientid('radiosnifferclient')

        staticfmt = {'site': site}
        _topics = mqttcommon.generate_topics_dict(
            self.SERVER.TOPICS, prefix, self.SERVER.AGENTTOPIC, staticfmt)

        _print_wrapper = self.async_print_handle_readlinebuff()
        error_cb = _print_wrapper(self.error_cb)
        raw_cb = _print_wrapper(self.raw_handler)
        self.topics = {
            'raw': mqttcommon.ChannelClient(_topics['noderaw'], raw_cb),
            'rawheader': mqttcommon.RequestClient(
                _topics['raw'], 'rawheader', clientid=clientid),
            'rawstart': mqttcommon.RequestClient(_topics['noderaw'], 'start',
                                                 clientid=clientid),

            'stop': mqttcommon.RequestClient(
                _topics['node'], 'stop', clientid=clientid),

            'stopall': mqttcommon.RequestClient(
                _topics['agenttopic'], 'stopall', clientid=clientid),

            'error': mqttcommon.ErrorClient(_topics['agenttopic'],
                                            callback=error_cb),
        }

        self.pcap_files = PcapFiles()
        self.client = client
        self.client.topics = list(self.topics.values())

    def error_cb(self, message, relative_topic):  # pylint:disable=no-self-use
        """Callback on 'error' topic."""
        msg = message.payload.decode('utf-8')
        print('RADIO SNIFFER ERROR: %s: %s' % (relative_topic, msg))

    @classmethod
    def from_opts_dict(cls, prefix, site, **kwargs):
        """Create class from argparse entries."""
        client = mqttcommon.MQTTClient.from_opts_dict(**kwargs)
        return cls(client, prefix=prefix, site=site)

    def do_rawpcap(self, arg):
        """Create pcap file and get RAW header."""
        try:
            pcapfd = open(arg, 'wb', buffering=0)
        except (IOError, OSError) as err:
            print('Could not open file: %s' % err)
            raise ValueError()

        topic = self.topics['rawheader']
        ret = topic.request(self.client, b'', timeout=5)

        # No error message
        header = ret

        print('Writing RAW PCAP header: %u bytes' % len(header))
        self.pcap_files.set('raw', pcapfd, header)

    def help_rawpcap(self):
        """Help rawpcap command."""
        print(self.RAWPCAP_USAGE, end='')

    # # # # # # # # #
    # rawpcapclose  #
    # # # # # # # # #
    def do_rawpcapclose(self, _):
        """Close rawpcap file."""
        self.pcap_files.clear('raw')

    def help_rawpcapclose(self):
        """Help rawpcapclose command."""
        print(self.RAWPCAPCLOSE_USAGE, end='')

    # # # #
    # raw #
    # # # #
    def raw_handler(self, message, archi, num):  # pylint:disable=no-self-use
        """Handle packets received from sniffer."""
        packet = message.payload

        message = 'PKT: len(%u)' % len(packet)
        print('raw_handler(%s-%s): %s' % (archi, num, message))

        self.pcap_files.write('raw', packet)

    def do_rawstart(self, arg):
        """Start sniffer on CHANNEL for given node: ARCHI NUM."""
        archi, num, channel = self.cmd_split(arg)
        num = int(num)
        channel = int(channel)

        self._do_rawstart(archi, num, channel)

    def _do_rawstart(self, archi, num, channel):
        topic = self.topics['rawstart']
        channel_str = str(channel).encode('utf-8')

        ret = topic.request(self.client, channel_str, timeout=5,
                            archi=archi, num=num)
        if ret:
            raise RuntimeError(ret.decode('utf-8'))

    def help_rawstart(self):
        """Help rawstart command."""
        print(self.RAWSTART_USAGE, end='')

    # # # # #
    # stop  #
    # # # # #
    def do_stop(self, arg):
        """Stop node sniffer: ARCHI NUM."""
        archi, num = self.cmd_split(arg)
        num = int(num)

        topic = self.topics['stop']
        ret = topic.request(self.client, b'', timeout=5,
                            archi=archi, num=num)
        if ret:
            raise RuntimeError(ret.decode('utf-8'))

    def help_stop(self):
        """Help stop command."""
        print(self.STOP_USAGE, end='')

    # # # # # #
    # stopall #
    # # # # # #
    def do_stopall(self, _):
        """Stop all nodes sniffer."""
        topic = self.topics['stopall']
        ret = topic.request(self.client, b'', timeout=5)
        if ret:
            raise RuntimeError(ret.decode('utf-8'))

    def help_stopall(self):
        """Help stopall command."""
        print(self.STOPALL_USAGE, end='')

    def run(self):
        """Run client and shell."""
        self.start()
        try:
            self.cmdloop()
        except KeyboardInterrupt:
            pass
        self.stop()

    def start(self):
        """Start Agent."""
        self.client.start()

    def stop(self):
        """Stop agent."""
        self.client.stop()


def main(opts=None):
    """SerialAgent shell client.

    :param opts: If provided, don't parse command line but use it instead
    :type opts: argparse Namespace object
    """
    opts = opts or PARSER.parse_args()
    shell = RadioSnifferShell.from_opts_dict(**vars(opts))
    shell.run()



# -*- coding:utf-8 -*-
"""Integration tests for IoT-LAB MQTT radiosniffer client and server."""
from __future__ import (absolute_import, division, print_function,
                        unicode_literals)
from builtins import *  # pylint:disable=W0401,W0614,W0622

import os
import contextlib
try:
    from StringIO import StringIO
except ImportError:
    from io import StringIO

import mock
from iotlabmqtt.clients import log as log_client

from . import IntegrationTestCase
from .. import TestCaseImproved


class LogIntegrationTest(IntegrationTestCase):
    """Test log client using a broker."""
    def setUp(self):
        self.prefix = 'log/test/prefix'
        self.timestamp = None
        self.client = None
        self.client_log_handler = None

    @contextlib.contextmanager
    def start_client(self, brokerport):
        """Start log client context manager.

        Yields client and stdout mock.
        """
        args = ['localhost', '--broker-port', '%s' % brokerport,
                '--prefix', self.prefix]
        opts = log_client.PARSER.parse_args(args)

        self.client_log_handler = log_client.LogShell.log_handler
        with mock.patch.object(log_client.LogShell, 'log_handler',
                               self._log_handler):
            self.client = log_client.LogShell.from_opts_dict(**vars(opts))
        self.client.start()

        try:
            with mock.patch('sys.stdout', new_callable=StringIO) as stdout:
                yield self.client, stdout
        finally:
            self.client.stop()

    def _log_handler(self, message):
        message.timestamp = self.timestamp
        self.client_log_handler(self.client, message)

    def _topic(self, topic):
        return os.path.join(self.prefix, topic)

    def test_log(self):
        """Test log cases."""
        with self.start_client(self.BROKERPORT) as (client, stdout):
            self._test_log(client, stdout)

    def _test_log(self, client, stdout):
        # Null
        self.timestamp = 1492008169.880
        message = (
            "1492008169.880 log/test/prefix/a/b/c ''\n"
            "(Cmd) "
        )
        client.client.publish(self._topic('a/b/c'), None)
        self.assertEqualTimeout(stdout.getvalue, message, 2)
        stdout.seek(0)
        stdout.truncate(0)

        # Null
        self.timestamp = 1492008946.123
        message = (
            "1492008946.123 log/test/prefix/null/b/c ''\n"
            "(Cmd) "
        )
        client.client.publish(self._topic('null/b/c'), b'')
        self.assertEqualTimeout(stdout.getvalue, message, 2)
        stdout.seek(0)
        stdout.truncate(0)

        # UTF8
        self.timestamp = 1492009044.880
        message = (
            "1492009044.880 log/test/prefix/u/ni/code 'Gaëtan'\n"
            "(Cmd) "
        )
        client.client.publish(self._topic('u/ni/code'),
                              'Gaëtan'.encode('utf-8'))
        self.assertEqualTimeout(stdout.getvalue, message, 2)
        stdout.seek(0)
        stdout.truncate(0)

        # bytes
        payload = bytearray(range(256))
        self.timestamp = 1492009308.999
        message = (
            "1492009308.999 log/test/prefix/b/ytes "
            "'BIN: (len 256): hash: 4916d6'\n"
            "(Cmd) "
        )
        client.client.publish(self._topic('b/ytes'), payload)
        self.assertEqualTimeout(stdout.getvalue, message, 2)
        stdout.seek(0)
        stdout.truncate(0)

    def test_log_and_outfile(self):
        """Test log cases with outfile."""
        with self.start_client(self.BROKERPORT) as (client, stdout):
            self._test_log_and_outfile(client, stdout)

    def _test_log_and_outfile(self, client, stdout):

        outfile = '/tmp/outlog'
        # Reuse _test_log to produce output

        # Nothing saved
        self._test_log(client, stdout)

        client.onecmd('open %s' % outfile)
        self._test_log(client, stdout)

        # Trim file and re-save
        client.onecmd('open %s' % outfile)
        self._test_log(client, stdout)

        # Close and no output save after
        client.onecmd('close')
        self._test_log(client, stdout)

        # Only one time
        output = open(outfile).read()
        expected = (
            '1492008169.880;log/test/prefix/a/b/c;\n'
            '1492008946.123;log/test/prefix/null/b/c;\n'
            '1492009044.880;log/test/prefix/u/ni/code;R2HDq3Rhbg==\n'
            '1492009308.999;log/test/prefix/b/ytes;'

            'AAECAwQFBgcICQoLDA0ODxAREhMUFRYXGBkaGxwdHh8gISIjJCUmJygpKissLS4vM'
            'DEyMzQ1Njc4OTo7PD0+P0BBQkNERUZHSElKS0xNTk9QUVJTVFVWV1hZWltcXV5fYG'
            'FiY2RlZmdoaWprbG1ub3BxcnN0dXZ3eHl6e3x9fn+AgYKDhIWGh4iJiouMjY6PkJG'
            'Sk5SVlpeYmZqbnJ2en6ChoqOkpaanqKmqq6ytrq+wsbKztLW2t7i5uru8vb6/wMHC'
            'w8TFxsfIycrLzM3Oz9DR0tPU1dbX2Nna29zd3t/g4eLj5OXm5+jp6uvs7e7v8PHy8'
            '/T19vf4+fr7/P3+/w==\n'
        )

        self.assertEqual(output, expected)

        os.remove(outfile)


@mock.patch('sys.stdout', new_callable=StringIO)
class LogClientErrorTests(TestCaseImproved):
    """Test Log parsing errors."""

    def setUp(self):
        args = ['localhost', '--broker-port', '%s' % 54321,
                '--prefix', 'log/test/prefix']
        opts = log_client.PARSER.parse_args(args)
        self.client = log_client.LogShell.from_opts_dict(**vars(opts))
        # Publish should crash
        self.client.publish = None

    def test_open(self, stdout):
        """Test open errors."""
        hlp = ('Error: Invalid arguments\n'
               'Usage: open LOGFILE\n'
               '  LOGFILE:   csv log file\n')

        self.client.onecmd('open')
        err = "Could not open file: [Errno 2] No such file or directory: ''\n"
        self.assertEqual(stdout.getvalue(), err + hlp)
        stdout.seek(0)
        stdout.truncate(0)

        # Cannot open file
        self.client.onecmd('open /non_existing/dir/and/file')
        err = ("Could not open file: [Errno 2] No such file or directory: "
               "'/non_existing/dir/and/file'\n")
        self.assertEqual(stdout.getvalue(), err + hlp)
        stdout.seek(0)

    def test_close(self, stdout):
        """Test close help."""
        hlp = ('close\n'
               '  Close current logfile\n')

        # cannot be done from command, just call it
        self.client.onecmd('help close')
        self.assertEqual(stdout.getvalue(), hlp)
        stdout.seek(0)
        stdout.truncate(0)



# -*- coding: utf-8 -*-

r"""IoT-LAB MQTT Radio Sniffer agent
====================================

Radio Sniffer Agent provides access to the IoT-LAB nodes radio sniffer.
Packets are encapsulated as ``pcap`` (`wireshark:pcap`_).

.. _wireshark\:pcap: https://wiki.wireshark.org/Development/LibpcapFileFormat

Radio Sniffer agent base topic: ::

   {prefix}/iot-lab/radiosniffer/{site}

Every topics from radio sniffer agent topics start by this topic prefix

:param prefix: configured prefix
:param site: site where the agent is run


Topics Summary
==============

.. |node|             replace::  ``{snifferagenttopic}/{archi}/{num}``
.. |rawheader|        replace::  ``{snifferagenttopic}/raw/ctl/header``
.. |raw|              replace::  |node|\ ``/raw``
.. |rawchannel|       replace::  |node|\ ``/raw/data``
.. |rawstart|         replace::  |node|\ ``/raw/ctl/start``
.. |rawstop|          replace::  |node|\ ``/raw/ctl/stop``
.. |stop|             replace::  |node|\ ``/ctl/stop``
.. |stopall|          replace::  ``{snifferagenttopic}/ctl/stopall``
.. |error_t|          replace::  ``{snifferagenttopic}/error/``

.. |channel|          replace::  :ref:`Channel <ChannelTopic>`
.. |request|          replace::  :ref:`Request <RequestTopic>`
.. |error|            replace::  :ref:`Error <ErrorTopic>`

.. |request_topic|    replace::  *{topic}*/**request/{clientid}/{requestid}**
.. |reply_topic|      replace::  *{topic}*/**reply/{clientid}/{requestid}**
.. |in_topic|         replace::  *{topic}*/**in**
.. |out_topic|        replace::  *{topic}*/**out**

+-+---------------------------------------------------------------+-----------+
| |Topic                                                          | Type      |
+=+===============================================================+===========+
|  **Radio sniffer agent**                                                    |
+-+---------------------------------------------------------------+-----------+
| ``{prefix}/iot-lab/radiosniffer/{site}``                                    |
+-+---------------------------------------------------------------+-----------+
| ||error_t|                                                      | |error|   |
+-+---------------------------------------------------------------+-----------+
| ||stopall|                                                      ||request|  |
+-+---------------------------------------------------------------+-----------+
|  **Node**                                                                   |
+-+---------------------------------------------------------------+-----------+
| ||stop|                                                         ||request|  |
+-+---------------------------------------------------------------+-----------+
|  **Raw packet sniffer**                                                     |
+-+---------------------------------------------------------------+-----------+
| ||rawheader|                                                    ||request|  |
+-+---------------------------------------------------------------+-----------+
| ||rawstart|                                                     ||request|  |
+-+---------------------------------------------------------------+-----------+
| ||rawstop|                                                      ||request|  |
+-+---------------------------------------------------------------+-----------+
| ||rawchannel|                                                   || Output   |
| |                                                               || |channel||
+-+---------------------------------------------------------------+-----------+


Radio sniffer Agent global topics
=================================

Error Topic
-----------

Asynchronous error messages are posted to error topic.
Failure on requests are not duplicated here.

* |error_t|


For format see: :ref:`ErrorTopic`


Stop all redirections
---------------------

Stop all started sniffer.

+-----------------------------------------------------------------------------+
| ``stopall`` request:                                                        |
|                                                                             |
+============+================================================================+
| Topic:     |    |stopall|                                                   |
+------------+-----------------------------------------+----------------------+
|**Message** | **Topic**                               | **Payload**          |
+------------+-----------------------------------------+----------------------+
| Request    | |request_topic|                         | *empty*              |
+------------+-----------------------------------------+----------------------+
| Reply      | |reply_topic|                           | *empty* or error_msg |
+------------+-----------------------------------------+----------------------+


Node topics
===========

.. _stop_sniffer:

Stop sniffer
------------

Stop given node sniffer.


+-----------------------------------------------------------------------------+
| ``stop`` request:                                                           |
+============+================================================================+
| Topic:     |    |stop|                                                      |
+------------+-----------------------------------------+----------------------+
|**Message** | **Topic**                               | **Payload**          |
+------------+-----------------------------------------+----------------------+
| Request    | |request_topic|                         | *empty*              |
+------------+-----------------------------------------+----------------------+
| Reply      | |reply_topic|                           | *empty* or error_msg |
+------------+-----------------------------------------+----------------------+


Raw packet sniffer
==================

Topics to access to the node sniffer in `raw` packet mode.
Sniffer listen to all radio packets on a given ``channel``.

It must first be started in ``raw`` mode to have the raw packet output.

Global pcap header must be queried independently first.

:param channel: 802.15.4 channel between 11 and 26

Start sniffer in *raw* mode
---------------------------

Start one node sniffer in *raw* mode on given ``channel``.

+-----------------------------------------------------------------------------+
| ``raw/start`` request:                                                      |
+============+================================================================+
| Topic:     |    |rawstart|                                                  |
+------------+-----------------------------------------+----------------------+
|**Message** | **Topic**                               | **Payload**          |
+------------+-----------------------------------------+----------------------+
| Request    | |request_topic|                         | ``Channel string``   |
+------------+-----------------------------------------+----------------------+
| Reply      | |reply_topic|                           | *empty* or error_msg |
+------------+-----------------------------------------+----------------------+


Stop raw sniffer
----------------

Equivalent to :ref:`stop_sniffer` adde here for completeness

+-----------------------------------------------------------------------------+
| ``raw/stop`` request:                                                       |
+============+================================================================+
| Topic:     |    |rawstop|                                                   |
+------------+-----------------------------------------+----------------------+
|**Message** | **Topic**                               | **Payload**          |
+------------+-----------------------------------------+----------------------+
| Request    | |request_topic|                         | *empty*              |
+------------+-----------------------------------------+----------------------+
| Reply      | |reply_topic|                           | *empty* or error_msg |
+------------+-----------------------------------------+----------------------+

RAW packet sniffer PCAP global header
-------------------------------------

PCAP format consists of two main parts, a global header, then messages with a
per message header: `wireshark:pcap`_

This request returns the global header for ``raw`` mode.


+-----------------------------------------------------------------------------+
| ``raw/header`` request:                                                     |
+============+================================================================+
| Topic:     |    |rawheader|                                                 |
+------------+-----------------------------------------+----------------------+
|**Message** | **Topic**                               | **Payload**          |
+------------+-----------------------------------------+----------------------+
| Request    | |request_topic|                         | *empty*              |
+------------+-----------------------------------------+----------------------+
| Reply      | |reply_topic|                           | `Global pcap header` |
+------------+-----------------------------------------+----------------------+


Raw packet sniffer
------------------

Sniffer sends 802.15.4 raw radio frames encapsulated as ``pcap``.

One sniffed packet is sent by message. PCAP payload format is::

   Packet PCAP Header | Packet data

PCAP Header contains: ::

   :4 bytes: Timestamp seconds
   :4 bytes: Timestamp microseconds
   :4 bytes: Number of octet saved
   :4 bytes: Actual lengt of packet


+-----------------------------------------------------------------------------+
| **802.15.4 RAW sniffer channel**                                            |
+============+================================================================+
| Topic:     |    |rawchannel|                                                |
+------------+-----------------------------------------+----------------------+
|**Message** | **Topic**                               | **Payload**          |
+------------+-----------------------------------------+----------------------+
| Output     | |out_topic|                             | PCAP packet          |
+------------+-----------------------------------------+----------------------+
"""

from __future__ import (absolute_import, division, print_function,
                        unicode_literals)
from builtins import *  # pylint:disable=W0401,W0614,W0622

import struct
import threading

from . import common
from . import iotlabapi
from . import mqttcommon
from . import asyncconnection

PARSER = common.MQTTAgentArgumentParser()
iotlabapi.parser_add_iotlabapi_args(PARSER)


class ZepToPcap(object):  # pylint:disable=too-few-public-methods
    """Zep to Pcap converter.

    On ``convert`` converts the message as a zep packet.
    """
    ZEP_PORT = 17754
    ZEP_HDR_LEN = 32
    ZEP_TIME_IDX = 9
    # http://www.tcpdump.org/linktypes.html

    LINKTYPE_ETHERNET = 1
    LINKTYPE_IEEE802_15_4 = 195  # with FCS

    # 1970 - 1900 in seconds
    NTP_JAN_1970 = 2208988800
    NTP_SECONDS_FRAC = 1 << 32

    # Network headers as network endian
    ETH_HDR = struct.pack(b'!3H3HH',
                          0, 0, 0,  # dst mac addr
                          0, 0, 0,  # src mac addr
                          0x0800)   # Protocol: (0x0800 == IP)

    LINKTYPES = {
        'RAW': LINKTYPE_IEEE802_15_4,
        'ZEP': LINKTYPE_ETHERNET,
    }

    def __init__(self, mode='RAW'):
        # configure "raw" mode
        # On raw, use linktype 802.15_4 else ethernet encapsulation
        self.mode = mode
        link_type = self.LINKTYPES[mode]
        self.header = self._pcap_main_header(link_type)

        _converter = {
            'RAW': self._convert_raw,
            'ZEP': self._convert_zep,
        }
        self.convert = _converter[mode]

    @staticmethod
    def _pcap_main_header(link_type):
        """ Return the main pcap file header for `link_type`

        PCAP headers as native endian
        """
        return struct.pack(
            b'=LHHLLLL',
            0xa1b2c3d4,  # Pcap header Little Endian
            2,           # File format major revision (i.e. pcap <2>.4)
            4,           # File format minor revision (i.e. pcap 2.<4>)
            0,           # GMT to local correction: 0 if timestamps are UTC
            0,           # accuracy of timestamps -> set it to 0
            0xffff,      # packet capture limit -> typically 65535
            link_type,   # Link (Ethernet/802.15.4 FCS/...)
        )

    @classmethod
    def _convert_raw(cls, packet):
        """ Only write the ZEP payload as pcap"""
        timestamp = cls._timestamp(packet)

        # extract payload from zep encapsulated data
        payload = packet[cls.ZEP_HDR_LEN:]

        # Only add pcap header
        length = len(payload)
        pcap_hdr = cls._pcap_header(length, timestamp[0], timestamp[1])
        length += len(pcap_hdr)

        # Actually write the data
        pcap = b''.join((pcap_hdr, payload))
        return pcap

    @classmethod
    def _convert_zep(cls, packet):
        """ Encapsulate ZEP data in pcap outfile """
        timestamp = cls._timestamp(packet)

        # Calculate all headers
        length = len(packet)

        udp_hdr = cls._udp_header(length)
        length += len(udp_hdr)

        ip_hdr = cls._ip_header(length)
        length += len(ip_hdr)

        eth_hdr = cls._eth_header(length)
        length += len(eth_hdr)

        pcap_hdr = cls._pcap_header(length, timestamp[0], timestamp[1])
        length += len(pcap_hdr)

        pcap = b''.join((pcap_hdr, eth_hdr, ip_hdr, udp_hdr, packet))
        return pcap

    @classmethod
    def _timestamp(cls, packet):
        """ Extract packet timestamp as an unix time tuple (s, us)
        Packet timestamp is in 'ntp' format.

        MSB are seconds stored since 1 january 1900
        LSB are fraction of seconds where 2**32 == 1 second
        """
        ntp_t = struct.unpack_from(b'!LL', packet, cls.ZEP_TIME_IDX)

        t_s = ntp_t[0] - cls.NTP_JAN_1970
        t_us = (1000000 * ntp_t[1]) // cls.NTP_SECONDS_FRAC

        return t_s, t_us

    @classmethod
    def _udp_header(cls, pkt_len):
        """ Get UDP Header

        2B - src_port: ZEP_PORT also but not required
        2B - dst_port: ZEP_PORT == 17754
        2B - length:   header + packet length
        2B - checksum: Disable == 0

        """
        hdr_struct = struct.Struct(b'!HHHH')
        udp_len = hdr_struct.size + pkt_len
        udp_hdr = hdr_struct.pack(cls.ZEP_PORT, cls.ZEP_PORT, udp_len, 0)
        return udp_hdr

    @classmethod
    def _ip_header(cls, pkt_len):
        """ Get the IP Header

        1B - Version | IHL:           0x45: [4 | 5] : [4b | 4b]
        1B - Type of Service:         0
        2B - Length:                  pkt_len + Header length
        2B - Identification:          0
        2B - Flags | Fragment offset: 0x4000: [2 | 0] : [3b | 13b]
        1B - TTL:                     0xff
        1B - Protocol:                0x11 UDP
        2B - Checksum:                calculated
        4B - Source Address:          0x7F000001 (127.0.0.1)
        4B - Destination Address:     0x7F000001 (127.0.0.1)

        """
        hdr_struct = struct.Struct(b'!BBHHHBBHLL')
        ip_len = hdr_struct.size + pkt_len

        # generate header with checksum == 0 to calculate checksum
        checksum = 0
        ip_hdr_csum = hdr_struct.pack(0x45, 0, ip_len, 0, 0x4000, 0xff, 0x11,
                                      checksum, 0x7F000001, 0x7F000001)
        checksum = cls._ip_checksum(ip_hdr_csum)

        # Generate header with correct checksum
        ip_hdr = hdr_struct.pack(0x45, 0, ip_len, 0, 0x4000, 0xff, 0x11,
                                 checksum, 0x7F000001, 0x7F000001)
        return ip_hdr

    @staticmethod
    def _ip_checksum(hdr):
        """ Calculate the ip checksum for given header """
        assert (len(hdr) % 2) == 0  # hdr has even length
        word_pack = struct.Struct(b'!H')

        # Sum all 16bits
        hdr_split = (hdr[i:i + 2] for i in range(0, len(hdr), 2))
        csum = sum((word_pack.unpack(word)[0] for word in hdr_split))

        # Reduce to 16b and save the one complement
        checksum = (csum + (csum >> 16)) & 0xFFFF ^ 0xFFFF
        return checksum

    @classmethod
    def _eth_header(cls, pkt_len):  # pylint:disable=unused-argument
        """ Return a static empty ethernet header

        6B - dst mac addr: 0
        6B - src mac addr: 0
        2B - protocol:     0x0800 (IP)
        """
        return cls.ETH_HDR

    @staticmethod
    def _pcap_header(pkt_len, t_s, t_us):
        """ Get the PCAP Header

        4B - Timestamp seconds:      current time
        4B - Timestamp microseconds: current time
        4B - Number of octet saved:  pkt_len
        4B - Actual lengt of packet: pkt_len
        """

        hdr_struct = struct.Struct(b'=LLLL')
        pcap_len = pkt_len
        pcap_hdr = hdr_struct.pack(t_s, t_us, pcap_len, pcap_len)
        return pcap_hdr


class ZEPHandler(object):  # pylint:disable=too-few-public-methods

    """ZEP data handler."""
    ZEP_HDR_LEN = 32  # zeptopcap.ZepPcap.ZEP_HDR_LEN
    ZEP_START = b'EX\2'

    def __init__(self, handler):
        self.data = b''
        self.handler = handler

    def __call__(self, input_data):
        """Call 'handler' on received data packet per packet."""
        data = self.data + input_data

        while True:
            data = self._strip_until_pkt_start(data)
            if (not data.startswith(self.ZEP_START) or
                    len(data) < self.ZEP_HDR_LEN):
                break

            # length = header length + data['len_byte']
            full_len = self.ZEP_HDR_LEN + self._payload_len(data)
            if len(data) < full_len:
                break

            # Extract packet
            pkt, data = data[:full_len], data[full_len:]
            self.handler(pkt)

        self.data = data

    def _payload_len(self, data):
        """Read 'length' byte from data.

        Length byte is byte at index ZEP_HDR_LEN -1.
        Use bytearray to be python2 and python3 compatible.
        `ord` does not work in python3.
        """
        return bytearray(data[self.ZEP_HDR_LEN - 1:self.ZEP_HDR_LEN])[0]

    @classmethod
    def _strip_until_pkt_start(cls, msg):
        """
        >>> msg = b'abcdEEEEEEEEEX\2'
        >>> b'EX\2' == ZEPHandler._strip_until_pkt_start(msg)
        True

        >>> msg = b'abcdEEEEEEEEEX\2' b'12345'
        >>> b'EX\2' b'12345' == ZEPHandler._strip_until_pkt_start(msg)
        True

        >>> msg = b'abcdEEE'
        >>> b'EE' == ZEPHandler._strip_until_pkt_start(msg)
        True

        >>> msg = b'abcdEEEa'
        >>> b'Ea' == ZEPHandler._strip_until_pkt_start(msg)
        True

        >>> msg = b'a'
        >>> b'a' == ZEPHandler._strip_until_pkt_start(msg)
        True

        """
        whole_index = msg.find(cls.ZEP_START)
        if whole_index == 0:   # is stripped
            return msg
        if whole_index != -1:  # found, strip first lines
            return msg[whole_index:]

        # not found but remove some chars from the buffer
        # at max 2 required in this case
        # might be invalid packet but keeps buffer small anymay
        return msg[-2:]


class SnifferConnection(asyncconnection.NodeConnection):
    """SnifferConnection implementation.

    Only overrides ``_address`` to adapt to the sniffer port.
    """
    PORT = 30000

    def _address(self, archi, num):  # overrides
        """Return socket address for archi/num.

        Hack for 'localhost' to use ``self.num`` as port.
        """
        # pylint:disable=no-else-return
        if archi == 'localhost':
            return archi, int(num)
        else:
            return ('%s-%s' % (archi, num), self.PORT)


class Node(object):  # pylint:disable=too-many-instance-attributes
    """Node resource managing sniffer connection, handling commands and events.

    :param archi: node architecture, or localhost
    :param num: node number
    :param closed_cb: callback for connection close
    :type closed_cb: Callable[[Node], None]
    :param error_cb: callback for asynchronous errors
    :type error_cb: Callable[[Node, str], None]
    """

    STATES = ('closed', 'startingsniffer', 'rawstarting', 'raw')
    CHANNELS = list(range(11, 26 + 1))

    def __init__(self, archi, num,  # pylint:disable=too-many-arguments
                 closed_cb, error_cb, iotlab_api, asyncoreservice=None):
        self.host = self.hostname(archi, num)
        self.closed_cb = closed_cb
        self.error_cb = error_cb

        self.channel = None
        self.iotlabapi = iotlab_api

        self.state = 'closed'
        self.reply_publisher = None
        self.connection = SnifferConnection(archi, num,
                                            self.conn_event_handler,
                                            service=asyncoreservice)

        # Required Rlock, connection socket errors calls event_handler('close')
        self._rlock = threading.RLock()

    @staticmethod
    def hostname(archi, num):
        """Return hostname for archi, num."""
        return archi, num

    @property
    def state(self):
        """Node state."""
        return self._state

    @state.setter
    def state(self, value):
        """Node state, check ``value`` is a valid state."""
        assert value in self.STATES
        self._state = value  # pylint:disable=attribute-defined-outside-init

    def _reply_request(self, reply, newstate=None):
        """Reply current request and remove publisher.

        If ``newstate`` is provided, update current state.
        """
        self.state = newstate or self.state
        self.reply_publisher(reply.encode('utf-8'))
        self.reply_publisher = None

    @common.synchronized('_rlock')
    def close(self):
        """Close connection and state."""
        self._close()
        return b''  # in case of a request

    @common.synchronized('_rlock')
    def conn_event_handler(self, event):
        """Handler for connection events."""
        event_handler = {
            'connect': self._event_connect,
            'error': self._event_error,
            'close': self._event_close,
        }
        return event_handler[event]()

    def _event_connect(self):
        if self.state == 'rawstarting':
            self._reply_request('', newstate='raw')
            return

        raise Exception('Got connect event in invalid state %s' % self.state)

    def _event_error(self):
        # Received other events exceptions
        error = common.traceback_error()
        previous_state = self._close()

        if previous_state == 'rawstarting':
            self._reply_request('Connection failed: %s' % error)
            return

        self.error_cb(self, error)

    def _event_close(self):
        # Jumps to event error
        raise Exception('Connection closed in state %s' % self.state)

    def _close(self):
        """Set 'closed' state, resets data_handler and call ``closed_cb``.

        Dont stop sniffer
        """
        previous_state = self.state
        self.connection.close()
        self.connection.data_handler = None
        self.state = 'closed'
        self.closed_cb(self)
        return previous_state

    @common.synchronized('_rlock')
    def req_rawstart(self, reply_publisher, zep_handler, channel):
        """Request to start sniffer redirection on ``channel``."""
        if self.state == 'raw':
            ret = 'Already started, stop before start to change channel'
            return ret.encode('utf-8')

        if self.state != 'closed':
            err = "Error: 'rawstarting' in mode %s. Wait or stop it first"
            return (err % self.state).encode('utf-8')

        self.state = 'startingsniffer'

        self.connection.data_handler = zep_handler
        self.reply_publisher = reply_publisher

        threading.Thread(target=self._thr_sniff_and_connect,
                         args=(channel,)).start()
        return None

    def _thr_sniff_and_connect(self, channel):
        # Should be run in a thread and not paho loop
        archi, num = self.host
        ret_dict = self.iotlabapi.set_sniffer_channel(channel, archi, num)

        profile_ret = ret_dict[str(num)]

        if profile_ret:
            with self._rlock:
                self.reply_publisher(profile_ret.encode('utf-8'))
                self.reply_publisher = None
                self._close()
        else:
            with self._rlock:
                self.state = 'rawstarting'
                self.connection.start()


class MQTTRadioSnifferAggregator(object):
    """Radio Sniffer Aggregator implementation for MQTT."""
    AGENTTOPIC = 'iot-lab/radiosniffer/{site}'
    TOPICS = {
        'raw': 'raw',
        'node': '{archi}/{num}',
        'noderaw': '{archi}/{num}/raw',
    }
    HOSTNAME = common.hostname()

    def __init__(self, client, prefix='', iotlab_api=None):
        assert iotlab_api
        super().__init__()

        staticfmt = {'site': self.HOSTNAME}
        _topics = mqttcommon.generate_topics_dict(self.TOPICS,
                                                  prefix, self.AGENTTOPIC,
                                                  staticfmt)

        self.topics = {
            'node': mqttcommon.NullTopic(_topics['node']),
            'rawheader': mqttcommon.RequestServer(
                _topics['raw'], 'rawheader', callback=self.cb_rawheader),

            'rawstart': mqttcommon.RequestServer(_topics['noderaw'], 'start',
                                                 callback=self.cb_rawstart),
            'rawstop': mqttcommon.RequestServer(_topics['noderaw'], 'stop',
                                                callback=self.cb_stop),
            'raw': mqttcommon.OutputChannelServer(_topics['noderaw']),

            'stop': mqttcommon.RequestServer(_topics['node'], 'stop',
                                             callback=self.cb_stop),

            'stopall': mqttcommon.RequestServer(_topics['agenttopic'],
                                                'stopall',
                                                callback=self.cb_stopall),

            'error': mqttcommon.ErrorServer(_topics['agenttopic']),
        }

        self.iotlabapi = iotlab_api
        self.nodes = {}
        self.asyncore = asyncconnection.AsyncoreService()

        self.client = client
        self.client.topics = list(self.topics.values())

    def error(self, topic, message):
        """Publish error that happend on topic."""
        self.topics['error'].publish_error(self.client, topic,
                                           message.encode('utf-8'))

    def _node_error(self, node, message):
        archi, num = node.host
        topic = self.topics['node'].topic.format(archi=archi, num=num)
        self.error(topic, message)

    def cb_rawstart(self, message, archi, num):
        """Start node sniffer in 'raw' mode.

        Create a new node if it does not currently exists.
        """
        try:
            channel = self._channel_from_payload(message.payload)
        except ValueError as err:
            return str(err).encode('utf-8')

        new_node = Node(archi, num, self._node_closed_cb, self._node_error,
                        self.iotlabapi, asyncoreservice=self.asyncore)
        node = self.nodes.setdefault(Node.hostname(archi, num), new_node)

        handler = self._raw_handler(archi, num)
        return node.req_rawstart(message.reply_publisher, handler, channel)

    @staticmethod
    def _channel_from_payload(payload):
        try:
            channel = int(payload.decode('utf-8'))
            if channel not in Node.CHANNELS:
                raise ValueError()
        except (TypeError, ValueError):
            raise ValueError('Invalid channel value, should be in [%s, %s]' %
                             (Node.CHANNELS[0], Node.CHANNELS[-1]))
        return channel

    def _raw_handler(self, archi, num):
        # """Raw handler for node ``archi``, ``num``.

        # Publish the message to the correct topic for ``archi``, ``num``.
        # """
        channel = self.topics['raw']
        publisher = channel.output_publisher(self.client, archi=archi, num=num)
        raw_encoder = ZepToPcap(mode='RAW').convert

        return ZEPHandler(lambda msg: publisher(raw_encoder(msg)))

    def _node_closed_cb(self, node):
        """Remove closed node."""
        self.nodes.pop(node.host, None)

    def cb_stop(self, message, archi, num):
        """Stop node redirection.

        In practice do not stop sniffer profile.
        """
        try:
            self.nodes[Node.hostname(archi, num)].close()
        except KeyError:
            pass
        return b''

    @staticmethod
    def cb_rawheader(_):
        """Return header for 'RAW' header."""
        return ZepToPcap(mode='RAW').header

    def cb_stopall(self, message):
        """Stop nodes sniffer redirection.

        In practice do not stop sniffer profile.
        """
        self._stop_all_nodes()
        return b''

    # Agent running

    def run(self):
        """Run agent."""
        try:
            self.start()
            common.wait_sigint()
        finally:
            self.stop()

    def start(self):
        """Start Agent."""
        self.asyncore.start()
        self.client.start()

    def stop(self):
        """Stop agent."""
        self.client.stop()
        self._stop_all_nodes()
        self.asyncore.stop()

    def _stop_all_nodes(self):
        """Close all nodes connections."""
        for node in list(self.nodes.values()):
            node.close()

    @classmethod
    def from_opts_dict(cls, prefix, **kwargs):
        """Create class from argparse entries."""
        api = iotlabapi.IoTLABAPI.from_opts_dict(**kwargs)
        client = mqttcommon.MQTTClient.from_opts_dict(**kwargs)
        return cls(client, prefix, iotlab_api=api)


def main():
    """Run radio sniffer agent."""
    opts = PARSER.parse_args()
    aggr = MQTTRadioSnifferAggregator.from_opts_dict(**vars(opts))
    aggr.run()



# -*- coding:utf-8 -*-
"""Tests for IoT-LAB MQTT"""

from __future__ import (absolute_import, division, print_function,
                        unicode_literals)

import time
import unittest


class TestCaseImproved(unittest.TestCase):
    """TestCase class with helper methods."""

    def assertEqualTimeout(self,  # pylint:disable=invalid-name
                           func, ret, timeout, step=0.1):
        """Assert func return equals ret before timeout."""
        end = time.time() + timeout
        while time.time() < end:
            if func() == ret:
                break
            time.sleep(step)
        self.assertEqual(func(), ret)



# -*- coding: utf-8 -*-

"""IoT-LAB API utils."""

from __future__ import (absolute_import, division, print_function,
                        unicode_literals)
from builtins import *  # pylint:disable=W0401,W0614,W0622

import tempfile

from . import common


def parser_add_iotlabapi_args(parser, group_help='IoT-LAB API configuration'):
    """Add iotlab auth arguments and experiment id to ``parser``."""
    group = parser.add_argument_group(group_help)
    group.add_argument('--iotlab-user', dest='iotlab_username',
                       help='IoT-LAB username')
    group.add_argument('--iotlab-password', dest='iotlab_password',
                       help='IoT-LAB password')
    group.add_argument('--experiment-id', dest='experiment_id', type=int,
                       help='experiment id submission')


class IoTLABAPI(object):
    """IoT-LAB API wrapper.

    Run commands using the REST API wrapped for the need of mqtt agents.
    It only runs commands for nodes on current execution site.
    """
    HOSTNAME = common.hostname()

    AUTH_ERROR = (
        'IoT-LAB API not authorized.\n'
        'Provide ``--iotlab-user`` and ``--iotlab-password`` options.\n'
        'Or register them using ``auth-cli --user USERNAME``'
    )

    def __init__(self, user=None, password=None, experiment_id=None):
        import iotlabcli.rest

        # Autoset user, password and experiment if they are None
        user, password = self._user_password(user, password)
        api = iotlabcli.rest.Api(user, password)
        # Get experiment ID and verify the experiment is Running.
        # This tests the user/password provided with the API
        experiment_id = self._experiment_id_and_running(api, experiment_id)

        self.api = api
        self.expid = experiment_id
        self.site = self.HOSTNAME

        assert self.expid is not None

    @staticmethod
    def _user_password(user, password):
        """Load username password from files if not defined."""
        import iotlabcli.auth
        return iotlabcli.auth.get_user_credentials(user, password)

    @classmethod
    def _experiment_id_and_running(cls, api, experiment_id=None):
        """Try getting experiment_id if None and check it is Running.

        System exit on failure.
        """
        import iotlabcli.rest
        try:
            experiment_id = cls._experiment_id(api, experiment_id)
            cls._check_experiment_running(api, experiment_id)
            return experiment_id
        except iotlabcli.rest.HTTPError:
            print(cls.AUTH_ERROR)
            exit(1)
        except (ValueError, RuntimeError) as err:
            print(err)
            exit(1)

    @staticmethod
    def _experiment_id(api, experiment_id):
        """Try getting experiment_id if not provided."""
        import iotlabcli.helpers
        return iotlabcli.helpers.get_current_experiment(api, experiment_id)

    @classmethod
    def _check_experiment_running(cls, api, experiment_id):
        """Check that experiment is Running."""
        ret = cls._get_experiment(api, experiment_id, 'state')
        state = ret['state']
        if state != 'Running':
            raise RuntimeError("Experiment %u not running '%s'" % (
                experiment_id, state))

    # Allow mocking this for tests
    @staticmethod
    def _get_experiment(api, experiment_id, info):
        """Get experiment ``info`` for ``experiment_id``."""
        import iotlabcli.experiment
        return iotlabcli.experiment.get_experiment(api, experiment_id, info)

    @classmethod
    def from_opts_dict(cls, iotlab_username=None, iotlab_password=None,
                       experiment_id=None, **_):
        """Create class from argparse entries."""
        return cls(iotlab_username, iotlab_password, experiment_id)

    def set_sniffer_channel(self, channel, archi, *nums):
        """Set sniffer on ``channel`` for nodes ``archi`` and ``*nums``."""
        return self._update_profile(channel, archi, *nums)

    def reset(self, archi, *nums):
        """Reset nodes ``archi`` and ``*nums``."""
        return self.node_command('reset', None, archi, *nums)

    def update_idle(self, archi, *nums):
        """Update nodes ``archi`` and ``*nums`` with idle firmware."""
        return self.node_command('update-idle', None, archi, *nums)

    def update(self, firmwarepath, archi, *nums):
        """Update nodes ``archi`` and ``*nums`` with ``firmwarepath``."""
        return self.node_command('update', firmwarepath, archi, *nums)

    def poweron(self, archi, *nums):
        """Power ON nodes ``archi`` and ``*nums``."""
        return self.node_command('start', None, archi, *nums)

    def poweroff(self, archi, *nums):
        """Power OFF nodes ``archi`` and ``*nums``."""
        return self.node_command('stop', None, archi, *nums)

    def _update_profile(self, channel, archi, *nums):
        """Update profile for nodes with archi and channel

        If ``archi`` == 'localhost' just ignore as used in tests.
        """
        if archi == 'localhost':
            return self.retval('', *nums)

        try:
            profile_str = self._sniffer_profile_json(archi, channel)
        except (ValueError, RuntimeError) as err:
            return self.retval('Create profile failed: %s' % err, *nums)

        with tempfile.NamedTemporaryFile() as prof:
            prof.write(profile_str.encode('ascii'))
            prof.flush()
            return self.node_command('profile-load', prof.name, archi, *nums)

    @staticmethod
    def retval(message, *nums):
        """Return value for ``message`` and ``*nums``."""
        return {str(num): message for num in nums}

    @staticmethod
    def _sniffer_profile_json(archi, channel):
        """Return sniffer profile json for ``archi`` and ``channel``.

        Raise ValueError or RuntimeError on error.
        """
        import iotlabcli.profile
        import iotlabcli.helpers
        profiles = {
            'm3': iotlabcli.profile.ProfileM3,
            'a8': iotlabcli.profile.ProfileA8,
        }

        name = 'iotlabmqtt_%u_%s' % (channel, archi)

        try:
            profile_class = profiles[archi]
        except KeyError:
            raise ValueError('Archi: %s not currently supported' % archi)

        profile = profile_class(name, 'dc')
        profile.set_radio('sniffer', [channel])

        return iotlabcli.helpers.json_dumps(profile)

    def node_command(self, command, cmd_opt, archi, *nums):
        """Run IoT-LAB node command and handle errors."""
        assert nums
        try:
            return self._node_command(command, cmd_opt, archi, *nums)
        except (IOError, RuntimeError) as err:
            msg = "IoT-LAB Request '%s' error: '%s'" % (command, err)
            return self.retval(msg, *nums)

    def _node_command(self, command, cmd_opt, archi, *nums):
        import iotlabcli.node
        nodes = self._nodes_for_num(archi, *nums)
        result = iotlabcli.node.node_command(self.api, command, self.expid,
                                             nodes, cmd_opt)
        return self._command_result_to_retval(result, archi)

    def _nodes_for_num(self, archi, *nums):
        """Return nodes address list for `archi` and `*nums`."""
        return [node_from_infos(archi, num, self.site) for num in nums]

    def _command_result_to_retval(self, ret_dict, archi):
        readable_value = {'0': '', '1': 'Execution failed on node'}
        result = {}
        for value, nodes_list in ret_dict.items():
            value = readable_value.get(value, value)
            for node in nodes_list:
                archi_, num, site_ = infos_from_node(node)
                numstr = str(num)
                assert (archi_, site_) == (archi, self.site)
                assert numstr not in result
                result[str(num)] = value
        return result


def node_from_infos(archi, num, site):  # pylint:disable=unused-argument
    """Node hostname from infos.

    >>> print(node_from_infos('m3', 1, 'grenoble'))
    m3-1.grenoble.iot-lab.info
    """
    fmt = '{archi}-{num}.{site}.iot-lab.info'
    return fmt.format(**locals())


def infos_from_node(node):
    """Infos from node hostname.

    >>> (infos_from_node('m3-1.grenoble.iot-lab.info') ==
    ...  ('m3', '1', 'grenoble'))
    True

    >>> (infos_from_node('node-a8-1.grenoble.iot-lab.info') ==
    ...  ('node-a8', '1', 'grenoble'))
    True
    """
    first, site = node.split('.', 2)[:2]
    archi, numstr = first.rsplit('-', 1)
    return (archi, numstr, site)



# -*- coding: utf-8 -*-

"""Clients for IoT-LAB MQTT agents"""

from __future__ import (absolute_import, division, print_function,
                        unicode_literals)

import argparse

from . import serial
from . import node
from . import radiosniffer
from . import process
from . import log


PARSER = argparse.ArgumentParser()
SUBPARSER = PARSER.add_subparsers(dest='command')
SUBPARSER.required = True

# Add parser that use given module parser
#   'add_help=False' prevents '--help' duplicate
SUBPARSER.add_parser('serial', parents=[serial.PARSER], add_help=False,
                     help='serial redirection client')
SUBPARSER.add_parser('node', parents=[node.PARSER],
                     add_help=False, help='node client')
SUBPARSER.add_parser('radiosniffer', parents=[radiosniffer.PARSER],
                     add_help=False, help='radiosniffer client')
SUBPARSER.add_parser('process', parents=[process.PARSER],
                     add_help=False, help='process client')
SUBPARSER.add_parser('log', parents=[log.PARSER],
                     add_help=False, help='log messages')


def main():
    """Run given client main function."""
    opts = PARSER.parse_args()
    client = globals()[opts.command]
    client.main(opts)



# -*- coding: utf-8 -*-

"""Processes management common."""

from __future__ import (absolute_import, division, print_function,
                        unicode_literals)
from builtins import *  # pylint:disable=W0401,W0614,W0622

try:
    import Queue as queue
except ImportError:  # pragma: no cover
    import queue
import os
import time
import threading
import subprocess
import functools

from . import common


def empty_queue(my_queue):
    """Empty message queue."""
    try:
        while True:
            my_queue.get_nowait()
    except queue.Empty:
        return


class Process(object):
    """Process manager."""
    STATES = ['New', 'Running', 'Stopping', 'Stopped']

    def __init__(self, closed_cb, stdout_cb, stderr_cb, error_cb):
        self.callbacks = {
            'closed': closed_cb,
            'stdout': stdout_cb,
            'stderr': stderr_cb,
            'error': error_cb,
        }
        self.state = 'New'
        self.command = None
        self.proc = None
        self.stdin_queue = queue.Queue(0)

        self._threads = {}
        self._rlock = threading.RLock()

    def is_idle(self):
        """Test if current process is not running."""
        return self.state in ('New', 'Stopped')

    @common.synchronized('_rlock')
    def run(self, command):
        """Run given `command`.

        Process should be in an idle state.
        """
        if not self.is_idle():
            return 'Process still running, be stopped before running'

        try:
            self.command = list(command)
            return self._run()
        except OSError as err:
            # Invalid command
            return '%s' % err
        except (TypeError, ValueError) as err:
            # Invalid arguments (maybe if not a list ?)
            return 'Invalid command'

    def _run(self):
        empty_queue(self.stdin_queue)
        self._popen()
        self.state = 'Running'
        return ''

    @common.synchronized('_rlock')
    def poll(self):
        """Check if process has terminated.

        Returns returncode string if there is one else '' or error.
        """
        if self.state == 'New':
            return 'Error process never run'

        if self.state != 'Stopped':
            return ''

        retcode = self.proc.returncode
        retcode = retcode if retcode is not None else ''
        return '%s' % retcode

    @common.synchronized('_rlock')
    def kill(self, timeout=1):
        """Kill current process."""
        if self.state != 'Running':  # Include 'Stopping'
            return ''

        self._kill(timeout=timeout)

        return ''

    def _kill(self, timeout=1):
        """Kill current process.

        Sends terminate then kill if still running after timeout
        """
        try:
            self.proc.terminate()
            self.proc_wait_timeout(self.proc, timeout)
            self.proc.kill()
        except OSError:
            pass

    @staticmethod
    def proc_wait_timeout(proc, timeout, step=0.05):
        """Wait for proc to terminate at max timeout."""
        t_end = time.time() + timeout
        while time.time() < t_end:
            if proc.poll() is not None:
                break
            time.sleep(step)
        return proc.poll()

    @common.synchronized('_rlock')
    def rerun(self):
        """Rerun current process."""
        if self.state == 'New':
            return 'Error process never run'
        if not self.is_idle():
            return 'Process still running, be stopped before running'

        return self._run()

    @common.synchronized('_rlock')
    def write(self, data):
        """Write data to process stdin."""
        if self.state == 'New':
            self.callbacks['error']('Process not running')
            return

        if self.state != 'Running':
            # Ignore here, no error as it could happen dynamically
            return

        self.stdin_queue.put(data)

    def _popen(self):
        # Set PYTHONUNBUFFERED as its our major language
        # And it will be used by agents
        env = os.environ.copy()
        env['PYTHONUNBUFFERED'] = '1'

        popen_kwargs = {
            'stdin': subprocess.PIPE,
            'stdout': subprocess.PIPE,
            'stderr': subprocess.PIPE,
            'shell': False,
            'close_fds': False,
            'env': env,
        }

        self.proc = subprocess.Popen(args=self.command, **popen_kwargs)
        self._threads['stdout'] = self._pipe_reader(
            'stdout', self.proc.stdout, self.callbacks['stdout'])
        self._threads['stderr'] = self._pipe_reader(
            'stderr', self.proc.stderr, self.callbacks['stderr'])
        self._threads['stdin'] = self._pipe_writer(
            'stdin', self.proc.stdin, self.stdin_queue)

    @common.synchronized('_rlock')
    def _handle_thread_end(self, name):
        """Handle current thread closing.

        On all thread closed, return.
        """
        self._threads.pop(name, None)
        if self._threads:
            self.state = 'Stopping'
            return

        # No more threads
        self.state = 'Stopped'
        self.callbacks['closed'](self.proc.returncode)

    def _pipe_reader(self, name, pipe, data_handler):
        """Reader from a pipe to data_handler."""
        args = (name, pipe, data_handler)
        thr = threading.Thread(target=self._pipe_reader_thr, args=args)
        thr.start()
        return thr

    def _pipe_reader_thr(self, name, pipe, data_handler, delay=0.1):
        """Thread target for reader from a pipe to data_handler."""
        self._set_nonblocking(pipe)
        while self._pipe_read_and_handle(pipe, data_handler, delay=delay):
            pass
        self._handle_thread_end(name)

    def _pipe_read_and_handle(self, pipe, data_handler, delay=0.1):
        """Try reading from pipe and handle data if any available.

        :returns: True if there can still be some data to read
        """
        try:
            data = pipe.read(1024)
        except IOError:
            data = ''

        if data:
            data_handler(data)
            return True

        if self.proc.poll() is None:
            time.sleep(delay)
            return True

        # No data and proc is dead
        # Send a last empty message
        data_handler(b'')
        return False

    @staticmethod
    def _set_nonblocking(pipe):
        """Set supbrocess pipe non blocking."""
        import fcntl
        flags = fcntl.fcntl(pipe, fcntl.F_GETFL)
        fcntl.fcntl(pipe, fcntl.F_SETFL, flags | os.O_NONBLOCK)

    def _pipe_writer(self, name, pipe, data_queue):
        """Writer to a pipe from data_queue."""
        args = (name, pipe, data_queue)
        thr = threading.Thread(target=self._pipe_writer_thr, args=args)
        thr.start()
        return thr

    def _pipe_writer_thr(self, name, pipe, data_queue, timeout=1):
        """Thread target for writer to a pipe from data_queue."""
        while self.proc.poll() is None:
            self._queue_get_and_write_pipe(pipe, data_queue, timeout=timeout)

        # Cleanup
        empty_queue(data_queue)

        self._handle_thread_end(name)

    @staticmethod
    def _queue_get_and_write_pipe(pipe, data_queue, timeout=1):
        """Get message from `data_queue` and write it to `pipe`."""
        try:
            data = data_queue.get(True, timeout)
            pipe.write(data)
            pipe.flush()
        except queue.Empty:
            pass
        except IOError:
            # write failed should be handled outside with proc.poll()
            pass


class ProcessManager(dict):
    """Process manager dict. Handles process by names."""

    def __init__(self, callbacks):
        super().__init__()
        self.process_callbacks = callbacks
        self._pid = 0
        self._rlock = threading.RLock()

    def start(self):
        """Start ProcessManager."""
        # Nothing to do currently
        pass

    def stop(self, timeout=5, step=0.05):
        """Stop ProcessManager.

        Kills all processes and wait for them to finish.
        """
        # First kill all before using timeout
        self._killall()

        t_end = time.time() + timeout
        while time.time() < t_end:
            if self._trystop():
                break
            time.sleep(step)

    def _trystop(self):
        """Try stopping all processes."""
        self._killall()
        return self._safefreeall()

    def _killall(self):
        """Kill all processes."""
        for proc in self.values():
            proc.kill()

    def _safefreeall(self):
        """Safe free all idle processes. Returns if all are freed."""
        for name in list(self.keys()):
            self.safefree(name)
        return not bool(self)

    @common.synchronized('_rlock')
    def new(self, name=''):
        """Allocate a new uniq process name.

        :param name: if provided, try using it, raise an error if already used
        """
        if not name:
            name = self._new_pid()
        name = str(name)

        if name in self:
            raise ValueError('name %s already in use' % name)

        self[name] = self._new_process(name)

        return name

    def _new_pid(self):
        # Increment to find a new PID
        while True:
            self._pid += 1
            pid = str(self._pid)
            if pid not in self:
                break

        return pid

    def _new_process(self, name):
        """Create a new process for `name`."""
        cb_with_name = [functools.partial(cb, name) for cb in
                        self.process_callbacks]
        return Process(*cb_with_name)

    @common.synchronized('_rlock')
    def free(self, name):
        """Free given process name process if idle."""
        proc = self[name]

        if not proc.is_idle():
            raise RuntimeError('Process still running wait or kill it first')

        self.pop(name)

    @common.synchronized('_rlock')
    def safefree(self, name):
        """Free given process name process if idle. Ignore if still running."""
        try:
            self.free(name)
        except RuntimeError:
            pass

    @common.synchronized('_rlock')
    def list(self):
        """List all processes."""
        return sorted(self.keys())
